
import numpy
import theano
import theano.sandbox.cuda
import theano.tensor as T
from theano.gof.opt import OpSub
from theano.compile import optdb
from theano import gof


def multi_batch_beam(array, start_idxs, batch_lens, beam_width, wrap_mode, pad_left=0, pad_right=0, idx_dim=0, batch_dim=1):
  """
  :param array: ndarray, at least 2D. symbolic
  :param start_idxs: ndarray, 1D. symbolic. can be float (for gpu)
  :param batch_lens: ndarray, 1D. symbolic. len of each batch. can be float (for gpu)
  :param beam_width: scalar. symbolic.
  :param wrap_mode: "wrap_around" or "pad". static.
  :param idx_dim: int. where to apply each start_idxs[i]. static.
  :param batch_dim: the same dim as in start_idxs. static.
  :param pad_value: used in wrap_mode "pad". automatically broadcasted. symbolic.
  :return: ndarray like array, but shape[idx_dim] == beam_width

  See also `_naive_multi_batch_beam` for one naive reference implementation.
  """
  assert array.ndim >= 2
  assert start_idxs.ndim == 1
  assert batch_lens.ndim == 1
  assert idx_dim < array.ndim
  assert batch_dim < array.ndim
  assert idx_dim != batch_dim
  assert wrap_mode in ("wrap_around", "pad")

  op = MultiBatchBeamOp(wrap_mode, idx_dim, batch_dim)
  return op(array, start_idxs, batch_lens, beam_width, pad_left, pad_right)


def _naive_multi_batch_beam(array, start_idxs, batch_lens, beam_width, wrap_mode, pad_left=0, pad_right=0, idx_dim=0, batch_dim=1):
  assert array.ndim >= 2
  assert start_idxs.ndim == 1
  assert batch_lens.ndim == 1
  assert idx_dim < array.ndim
  assert batch_dim < array.ndim
  assert idx_dim != batch_dim
  n_batch = array.shape[batch_dim]
  assert start_idxs.shape == (n_batch, )
  assert batch_lens.shape == (n_batch, )
  pad_left = numpy.asarray(pad_left)
  pad_left_bc = pad_left.reshape(*([1] * (array.ndim - pad_left.ndim) + list(pad_left.shape)))
  pad_right = numpy.asarray(pad_right)
  pad_right_bc = pad_right.reshape(*([1] * (array.ndim - pad_right.ndim) + list(pad_right.shape)))

  if idx_dim != 0: raise NotImplementedError  # This is usually the time dim.
  if batch_dim != 1: raise NotImplementedError
  # Thus, array is usually in format (time,batch,dim).

  beam = numpy.zeros((beam_width, n_batch) + array.shape[2:], dtype=array.dtype)
  for i0 in range(beam_width):
    for i1 in range(n_batch):
      idx = start_idxs[i1] + i0
      if wrap_mode == "wrap_around":
        idx = idx % batch_lens[i1]
      elif wrap_mode == "pad":
        if idx < 0:
          beam[i0, i1] = pad_left_bc
          continue
        elif idx >= batch_lens[i1]:
          beam[i0, i1] = pad_right_bc
          continue
      beam[i0, i1] = array[idx, i1]
  return beam


def _theano_cpu_multi_batch_beam(array, start_idxs, batch_lens, beam_width, wrap_mode, pad_left=0, pad_right=0, idx_dim=0, batch_dim=1):
  array = T.as_tensor(array)
  start_idxs = T.as_tensor(start_idxs)
  if start_idxs.dtype.startswith("float"):
    start_idxs = T.iround(start_idxs)
  batch_lens = T.as_tensor(batch_lens)
  if batch_lens.dtype.startswith("float"):
    batch_lens = T.iround(batch_lens)
  beam_width = T.as_tensor(beam_width)
  if beam_width.dtype.startswith("float"):
    beam_width = T.iround(beam_width)
  pad_left = T.as_tensor(pad_left)
  pad_right = T.as_tensor(pad_right)
  assert array.ndim >= 2
  assert start_idxs.ndim == 1
  assert batch_lens.ndim == 1
  assert beam_width.ndim == 0
  assert idx_dim < array.ndim
  assert batch_dim < array.ndim
  assert idx_dim != batch_dim
  n_batch = array.shape[batch_dim]

  if idx_dim != 0: raise NotImplementedError
  if batch_dim != 1: raise NotImplementedError
  if wrap_mode != "wrap_around": raise NotImplementedError

  idxs_0 = start_idxs.dimshuffle('x', 0)  # (beam,batch)
  idxs = idxs_0 + T.arange(beam_width).dimshuffle(0, 'x')  # (beam,batch)
  idxs_wrapped = idxs % batch_lens.dimshuffle('x', 0)  # (beam,batch)
  batches = T.arange(n_batch)  # (batch,)
  beam = array[idxs_wrapped[:, batches], batches]  # (beam,batch,...)
  if wrap_mode == "wrap_around":
    pass  # Done that.
  elif wrap_mode == "pad":
    cond_left = T.lt(idxs, 0)  # (beam,batch)
    cond_right = T.ge(idxs, batch_lens.dimshuffle('x', 0))  # (beam,batch)
    cond_left_bc = cond_left.dimshuffle(beam_width, n_batch, *([1] * (array.ndim - 2)))
    cond_right_bc = cond_right.dimshuffle(beam_width, n_batch, *([1] * (array.ndim - 2)))
    pad_left_bc = pad_left.dimshuffle(*(['x'] * (array.ndim - pad_left.ndim) +
                                        [pad_left.shape[i] for i in range(pad_left.ndim)]))
    pad_right_bc = pad_left.dimshuffle(*(['x'] * (array.ndim - pad_right.ndim) +
                                         [pad_right.shape[i] for i in range(pad_right.ndim)]))
    beam = T.switch(cond_left_bc, beam, T.cast(pad_left_bc, dtype=array.dtype))
    beam = T.switch(cond_right_bc, beam, T.cast(pad_right_bc, dtype=array.dtype))
  else:
    raise Exception("MultiBatchBeam: unknown wrap mode: %r" % wrap_mode)
  return beam


def _naive_multi_batch_beam_grad(array, start_idxs, batch_lens, beam_width, wrap_mode, pad_left=0, pad_right=0, idx_dim=0, batch_dim=1, output_grad=None):
  assert array.ndim >= 2
  assert start_idxs.ndim == 1
  assert batch_lens.ndim == 1
  assert idx_dim < array.ndim
  assert batch_dim < array.ndim
  assert idx_dim != batch_dim
  n_batch = array.shape[batch_dim]
  assert start_idxs.shape == (n_batch, )
  assert batch_lens.shape == (n_batch, )
  D_beam = output_grad
  pad_left = numpy.asarray(pad_left)
  pad_right = numpy.asarray(pad_right)

  if idx_dim != 0: raise NotImplementedError  # This is usually the time dim.
  if batch_dim != 1: raise NotImplementedError
  assert D_beam.shape == (beam_width, n_batch) + array.shape[2:]
  # Thus, array is usually in format (time,batch,dim).

  D_array = numpy.zeros_like(array, dtype="float32")
  if wrap_mode == "pad":
    D_pad_left = numpy.zeros(array.shape[2:], dtype="float32")
    D_pad_right = numpy.zeros(array.shape[2:], dtype="float32")
  else:
    D_pad_left = D_pad_right = None

  for i0 in range(beam_width):
    for i1 in range(n_batch):
      idx = start_idxs[i1] + i0
      if wrap_mode == "wrap_around":
        idx = idx % batch_lens[i1]
      elif wrap_mode == "pad":
        if idx < 0:
          D_pad_left += D_beam[i0, i1]
          continue
        if idx >= batch_lens[i1]:
          D_pad_right += D_beam[i0, i1]
          continue
      D_array[idx, i1] = D_beam[i0, i1]

  if wrap_mode == "pad":
    if D_pad_left.ndim > pad_left.ndim:
      D_pad_left = numpy.sum(D_pad_left, axis=tuple(range(D_pad_left.ndim - pad_left.ndim)))
    if D_pad_right.ndim > pad_right.ndim:
      D_pad_right = numpy.sum(D_pad_right, axis=tuple(range(D_pad_right.ndim - pad_right.ndim)))
    assert D_pad_left.shape == pad_left.shape
    assert D_pad_right.shape == pad_right.shape
  return D_array, D_pad_left, D_pad_right


def _theano_cpu_multi_batch_beam_grad(array, start_idxs, batch_lens, beam_width, wrap_mode, pad_left=0, pad_right=0, idx_dim=0, batch_dim=1, output_grad=None):
  # Note: This is slow and hacky. This will create an index-array of the size of the original array.
  # This is calculated on the CPU. The subtensor then can be done on the GPU, but we should avoid the first part.
  D_beam = output_grad
  prod_array_shape = T.prod(array.shape)
  prod_pad_left_shape = T.prod(pad_left.shape)
  prod_pad_right_shape = T.prod(pad_right.shape)
  D_array_tmp_size = prod_array_shape
  if wrap_mode == "pad":
    D_array_tmp_size += prod_pad_left_shape + prod_pad_right_shape
  D_array_tmp_flat = T.zeros([D_array_tmp_size], dtype="float32")  # with pad values
  if wrap_mode == "pad":
    # Calculate the indices for D_pad_left/D_pad_right in D_array_tmp_flat.
    pad_left_idxs = T.arange(prod_pad_left_shape) + prod_array_shape
    pad_right_idxs = T.arange(prod_pad_right_shape) + prod_array_shape + prod_pad_left_shape
    pad_left_idxs = pad_left_idxs.reshape(pad_left.shape)
    pad_right_idxs = pad_right_idxs.reshape(pad_right.shape)
  else:
    pad_left_idxs = pad_right_idxs = 0
  all_idxs = T.arange(T.prod(array.shape)).reshape(array.shape)
  idxs = multi_batch_beam(array=all_idxs, start_idxs=start_idxs, batch_lens=batch_lens, beam_width=beam_width,
                          wrap_mode=wrap_mode,
                          pad_left=pad_left_idxs, pad_right=pad_right_idxs,
                          idx_dim=idx_dim, batch_dim=batch_dim)
  D_array_tmp_flat = T.inc_subtensor(D_array_tmp_flat[idxs.flatten()], D_beam.flatten())
  if wrap_mode == "pad":
    D_array = D_array_tmp_flat[:prod_array_shape].reshape(array.shape)
    D_pad_left = D_array_tmp_flat[pad_left_idxs.flatten()].reshape(pad_left.shape)
    D_pad_right = D_array_tmp_flat[pad_right_idxs.flatten()].reshape(pad_right.shape)
  else:
    D_array = D_array_tmp_flat.reshape(array.shape)
    D_pad_left = D_pad_right = T.DisconnectedType()()

  return D_array, D_pad_left, D_pad_right


class MultiBatchBeamOp(theano.Op):
  __props__ = ("wrap_mode", "idx_dim", "batch_dim")

  def __init__(self, wrap_mode, idx_dim=0, batch_dim=1):
    super(MultiBatchBeamOp, self).__init__()
    self.wrap_mode = wrap_mode
    self.idx_dim = idx_dim
    self.batch_dim = batch_dim

  def make_node(self, array, start_idxs, batch_lens, beam_width, pad_left, pad_right):
    array = T.as_tensor_variable(array)
    start_idxs = T.as_tensor_variable(start_idxs)
    batch_lens = T.as_tensor_variable(batch_lens)
    beam_width = T.as_tensor_variable(beam_width)
    pad_left = T.as_tensor_variable(pad_left)
    pad_right = T.as_tensor_variable(pad_right)
    assert array.ndim >= 2
    assert start_idxs.ndim == 1
    assert batch_lens.ndim == 1
    assert beam_width.ndim == 0
    return theano.Apply(self, [array, start_idxs, batch_lens, beam_width, pad_left, pad_right], [array.type()])

  def perform(self, node, inputs, output_storage):
    array, start_idxs, batch_lens, beam_width, pad_left, pad_right = inputs
    beam_out, = output_storage
    n_batches = array.shape[self.batch_dim]
    assert start_idxs.shape[0] == n_batches
    assert batch_lens.shape[0] == n_batches
    if not start_idxs.dtype.name.startswith("int"):
      start_idxs = numpy.round(start_idxs).astype("int64")
    if not batch_lens.dtype.name.startswith("int"):
      batch_lens = numpy.round(batch_lens).astype("int64")
    if not beam_width.dtype.name.startswith("int"):
      beam_width = int(numpy.round(beam_width))
    pad_left_bc = numpy.asarray(pad_left).reshape(*([1] * (array.ndim - pad_left.ndim) + list(pad_left.shape)))
    pad_right_bc = numpy.asarray(pad_right).reshape(*([1] * (array.ndim - pad_right.ndim) + list(pad_right.shape)))

    idxs_bc = numpy.arange(beam_width).reshape(beam_width, 1)  # dimshuffle(0, 'x')  (beam,batch)
    start_idxs_bc = start_idxs.reshape(1, n_batches)  # dimshuffle('x', 0)  (beam,batch)
    idxs = idxs_bc + start_idxs_bc  # (beam,batch)
    batch_lens_bc = batch_lens.reshape(1, n_batches)  # dimshuffle('x', 0)  (beam,batch)
    idxs_wrapped = idxs % batch_lens_bc
    array_remaining_dims = sorted(set(range(array.ndim)) - set([self.idx_dim, self.batch_dim]))
    array_trans_dims_order = [self.idx_dim, self.batch_dim] + array_remaining_dims
    array_trans = array.transpose(*array_trans_dims_order)  # (time,batch,...)
    beam_trans = array_trans[idxs_wrapped, numpy.arange(n_batches)]  # (beam,batch,...)
    if self.wrap_mode == "wrap_around":
      pass  # We have done exactly that.
    elif self.wrap_mode == "pad":
      cond_left = idxs < 0  # (beam,batch)
      cond_right = idxs >= batch_lens_bc  # (beam,batch)
      cond_left_bc = cond_left.reshape(beam_width, n_batches, *([1] * len(array_remaining_dims)))
      cond_right_bc = cond_right.reshape(beam_width, n_batches, *([1] * len(array_remaining_dims)))
      beam_trans = numpy.where(cond_left_bc, numpy.cast[array.dtype](pad_left_bc), beam_trans)
      beam_trans = numpy.where(cond_right_bc, numpy.cast[array.dtype](pad_right_bc), beam_trans)
    else:
      raise Exception("MultiBatchBeam: unknown wrap mode: %r" % self.wrap_mode)
    beam = beam_trans.transpose(*map(array_trans_dims_order.index, range(array.ndim)))
    beam_out[0] = beam

  def infer_shape(self, node, input_shapes):
    array, start_idxs, batch_lens, beam_width, pad_left, pad_right = node.inputs
    beam_width = T.cast(beam_width, dtype="int64")
    array_shape, start_idxs_shape, batch_lens_shape, beam_width_shape, pad_left_shape, pad_right_shape = input_shapes
    beam_shape = [beam_width if i == self.idx_dim else array_shape[i] for i in range(len(array_shape))]
    return [tuple(beam_shape)]

  def grad(self, inputs, output_grads):
    array, start_idxs, batch_lens, beam_width, pad_left, pad_right = inputs
    D_beam, = output_grads

    if not isinstance(pad_left, theano.Constant):
      raise NotImplementedError("D_pad_left not implemented...")
    if not isinstance(pad_right, theano.Constant):
      raise NotImplementedError("D_pad_right not implemented...")

    grad_op = MultiBatchBeamGradAddOp(wrap_mode=self.wrap_mode, zero_with_shape=True, array_ndim=array.ndim,
                                      idx_dim=self.idx_dim, batch_dim=self.batch_dim)
    D_array = grad_op(array.shape, start_idxs, batch_lens, beam_width, D_beam)

    if self.wrap_mode == "wrap_around":
      D_pad_left = D_pad_right = T.DisconnectedType()()
    elif self.wrap_mode == "pad":
      D_pad_left = D_pad_right = T.DisconnectedType()()
      # XXX...
      # D_pad_left = T.zeros(pad_left.shape, dtype="float32")
      # D_pad_right = T.zeros(pad_right.shape, dtype="float32")
    else:
      assert False, self.wrap_mode

    # Those are all discrete values. The gradient is 0 almost everywhere, except for integers where it is not defined.
    D_start_idxs = T.DisconnectedType()()
    D_batch_lens = T.DisconnectedType()()
    D_beam_width = T.DisconnectedType()()

    return [D_array, D_start_idxs, D_batch_lens, D_beam_width, D_pad_left, D_pad_right]

  def connection_pattern(self, node):
    # Only the gradient of the first input (array) will be connected.
    # All others are disconnected (because round() or floor() is used on them.).
    pattern = [[True], [False], [False], [False], [False], [False]]
    # if self.wrap_mode == "pad":  # XXX... we assume constant for now
    #   pattern[-2:] = [[True], [True]]
    assert len(pattern) == len(node.inputs)
    return pattern


def _len_of_shape(shape):
  if isinstance(shape, (list,tuple)):
    return len(shape)
  if isinstance(shape, T.Apply):
    if isinstance(shape.op, T.Shape):
      assert len(shape.inputs) == 1
      return shape.inputs[0].ndim
  raise NotImplementedError("cannot handle %r" % shape)


inplace_increment = None
if theano.config.cxx:
  import theano.gof.cutils  # needed to import cutils_ext
  try:
    from cutils_ext.cutils_ext import inplace_increment
  except ImportError:
    pass


class MultiBatchBeamGradAddOp(theano.Op):
  __props__ = ("wrap_mode", "idx_dim", "batch_dim", "inplace", "zero_with_shape", "array_ndim")

  def __init__(self, wrap_mode, idx_dim=0, batch_dim=1, inplace=False, zero_with_shape=False, array_ndim=None):
    """
    (D_array / D_array_shape, start_idxs, batch_lens, beam_width, D_beam) -> D_array + grad

    :param str wrap_mode: "wrap_around" or "pad"
    :param int idx_dim: usually that's time dim
    :param int batch_dim: batch dim
    :param bool inplace: operate inplace on input
    :param bool zero_with_shape: we get D_array_shape as the first input and init D_array with zero
    :param int array_ndim: ndim of array/D_array. needed for zero_with_shape
    """
    super(MultiBatchBeamGradAddOp, self).__init__()
    self.wrap_mode = wrap_mode
    self.idx_dim = idx_dim
    self.batch_dim = batch_dim
    self.inplace = inplace
    self.zero_with_shape = zero_with_shape
    self.array_ndim = array_ndim
    if zero_with_shape:
      assert not inplace
      assert array_ndim > 0
    if inplace:
      # We operate inplace on D_array.
      self.destroy_map = {0: [0]}

  def make_node(self, D_array_or_shape, start_idxs, batch_lens, beam_width, D_beam):
    # XXX: Currently without D_pad_left and D_pad_right.
    start_idxs = T.as_tensor_variable(start_idxs)
    batch_lens = T.as_tensor_variable(batch_lens)
    beam_width = T.as_tensor_variable(beam_width)
    D_beam = T.as_tensor_variable(D_beam)
    if self.zero_with_shape:
      D_array_ndim = self.array_ndim
    else:
      D_array_ndim = D_array_or_shape.ndim
    assert start_idxs.ndim == 1
    assert batch_lens.ndim == 1
    assert beam_width.ndim == 0
    return theano.Apply(self,
                        [D_array_or_shape, start_idxs, batch_lens, beam_width, D_beam],
                        [T.TensorType("float32", (False,) * D_array_ndim)("D_array")])

  def infer_shape(self, node, input_shapes):
    if self.zero_with_shape:
      D_array_ndim = self.array_ndim
      shape = node.inputs[0]  # This is symbolic.
      return [[shape[i] for i in range(D_array_ndim)]]
    else:
      return [input_shapes[0]]

  def perform(self, node, inputs, output_storage):
    D_array_or_shape, start_idxs, batch_lens, beam_width, D_beam = inputs
    out_D_array, = output_storage
    if self.inplace:
      out_D_array[0] = D_array = D_array_or_shape
    elif self.zero_with_shape:
      out_D_array[0] = D_array = numpy.zeros(D_array_or_shape, "float32")
    else:
      out_D_array[0] = D_array = D_array_or_shape.copy()
    n_batches = D_array.shape[self.batch_dim]

    idxs_bc = numpy.arange(beam_width).reshape(beam_width, 1)  # dimshuffle(0, 'x')  (beam,batch)
    start_idxs_bc = start_idxs.reshape(1, n_batches)  # dimshuffle('x', 0)  (beam,batch)
    idxs = idxs_bc + start_idxs_bc  # (beam,batch)
    batch_lens_bc = batch_lens.reshape(1, n_batches)  # dimshuffle('x', 0)  (beam,batch)
    assert idxs.shape == D_beam.shape[:2]

    idxs = idxs.astype("int32")
    if self.wrap_mode == "wrap_around":
      idxs = idxs % batch_lens_bc.astype("int32")
    elif self.wrap_mode == "pad":
      idxs = numpy.where(idxs >= batch_lens_bc, -1, idxs)
      cond_bc = (idxs < 0).reshape(*(D_beam.shape[:2] + (1,) * (D_beam.ndim - 2)))
      D_beam = numpy.where(cond_bc, numpy.float32(0), D_beam)  # XXX: ignore padding part
      idxs = numpy.where(idxs < 0, 0, idxs)
    else:
      assert False, self.wrap_mode

    if self.idx_dim != 0: raise NotImplementedError  # TODO...
    if self.batch_dim != 1: raise NotImplementedError  # TODO...
    # In Numpy, x[idx] += y doesn't work if the same index is present
    # many times: it does it only once. Is it a bug? In any case, for
    # this reason we implement our own 'inc' iteration.
    # See also theano.tensor.subtensor.AdvancedIncSubtensor documentation.
    if inplace_increment is None: raise NotImplementedError("need Numpy 1.8 or later")
    # This is like D_array_and_pad[idxs, numpy.arange(n_batches)] += D_beam .
    inplace_increment(D_array, (idxs, numpy.arange(n_batches)), D_beam)


# https://deeplearning.net/software/theano/extending/optimization.html
# See also theano/compile/mode.py for reference about the position priority numbers.
# After priority 50 we can do destructive inplace operations.

@gof.local_optimizer([T.add])
def add_merge_MultiBatchBeamGradAddOp(node):
  if node.op != T.add: return False
  if len(node.inputs) < 2: return False
  grad_op_idx = None
  grad_op_v = None
  grad_op = None
  for i, input in enumerate(node.inputs):
    if input.owner and isinstance(input.owner.op, MultiBatchBeamGradAddOp):
      grad_op = input.owner.op
      if not grad_op.inplace:  # we cannot merge when we operate inplace on it
        grad_op_v = input
        grad_op_idx = i
        break
  if grad_op_idx is None: return False
  sum_inputs = [node.inputs[i] for i in range(len(node.inputs)) if i != grad_op_idx]
  if grad_op.zero_with_shape:
    # Make new grad_op without zero_with_shape.
    kwargs = {k: getattr(grad_op, k) for k in grad_op.__props__}
    kwargs["zero_with_shape"] = False
    grad_op = grad_op.__class__(**kwargs)
  else:
    old_grad_op_input0 = grad_op_v.owner.inputs[0]
    sum_inputs = [old_grad_op_input0] + sum_inputs
  assert len(sum_inputs) > 0
  if len(sum_inputs) == 1:
    new_grad_op_input0 = sum_inputs[0]
  else:
    new_grad_op_input0 = T.add(*sum_inputs)
  new_grad_op_inputs = [new_grad_op_input0] + grad_op_v.owner.inputs[1:]
  new_v = grad_op(*new_grad_op_inputs)
  return [new_v]

optdb.register('add_merge_MultiBatchBeamGradAddOp',
               gof.TopoOptimizer(add_merge_MultiBatchBeamGradAddOp),
               0.1, 'fast_run')


@gof.local_optimizer([MultiBatchBeamGradAddOp], inplace=True)
def inplace_MultiBatchBeamGradAddOp(node):
  if isinstance(node.op, MultiBatchBeamGradAddOp) and not node.op.inplace and not node.op.zero_with_shape:
    kwargs = {k: getattr(node.op, k) for k in node.op.__props__}
    kwargs["inplace"] = True
    new_op = node.op.__class__(**kwargs)
    new_v = new_op(*node.inputs)
    return [new_v]
  return False

optdb.register('inplace_MultiBatchBeamGradAddOp',
               gof.TopoOptimizer(inplace_MultiBatchBeamGradAddOp
                                 , failure_callback=gof.TopoOptimizer.warn_inplace
                                 ),
               76,  # after ScanInplaceOptimizer
               'fast_run', 'inplace')


class GpuMultiBatchBeamOp(theano.sandbox.cuda.GpuOp):
  pass  # TODO




from __future__ import print_function

import _setup_test_env  # noqa

import logging
import sys
import os
import os.path
import tensorflow as tf
import numpy as np
import unittest
import contextlib

from returnn.tf.network import *
from returnn.tf.layers.signal_processing import *
from returnn.config import Config

from returnn.util import better_exchook
better_exchook.replace_traceback_format_tb()


@contextlib.contextmanager
def make_scope():
  """
  :rtype: tf.compat.v1.Session
  """
  with tf.Graph().as_default() as graph:
    with tf_compat.v1.Session(graph=graph) as session:
      yield session


def test_melFilterbankLayer():
  with make_scope() as session:
    n_in, n_out = 257, 3
    layer_name = "mel_filterbank_layer"
    config = Config()
    config.update({
      "num_outputs": n_out,
      "num_inputs": n_in,
      "network": {
        layer_name: {
          "class": "mel_filterbank", "fft_size": 512, "nr_of_filters": n_out, "n_out": n_out, "is_output_layer": True,
          "from": "data:data"}
      }})
    network = TFNetwork(config=config, train_flag=True)
    network.construct_from_dict(config.typed_value("network"))
    layer = network.layers[layer_name]
    test_out = session.run(
      layer.output.placeholder,
      feed_dict={network.get_extern_data('data').placeholder: np.ones((1, 1, 257))})
    assert np.sum(test_out - np.asarray([28.27923584, 53.10634232, 99.71585846], dtype=np.float32)) < 1e-5


def test_complexLinearProjectionLayer():
  with make_scope() as session:
    n_in, n_out = 514, 128
    layer_name = "clp_layer"
    config = Config()
    config.update({
      "num_outputs": n_out,
      "num_inputs": n_in,
      "network": {
        layer_name: {
          "class": "complex_linear_projection", "nr_of_filters": n_out, "n_out": n_out, "is_output_layer": True,
          "from": "data:data"}
      }})
    network = TFNetwork(config=config, train_flag=True)
    network.construct_from_dict(config.typed_value("network"))
    layer = network.layers[layer_name]
    assert isinstance(layer, ComplexLinearProjectionLayer)
    i_r = np.ones((1, n_in // 2))
    i_i = np.ones((1, n_in // 2)) * 0.5
    test_input = np.expand_dims(np.reshape(np.transpose(
      np.reshape(np.concatenate([i_r, i_i], axis=1), (1, 2, 257)), [0, 2, 1]), (1, 514)), 0)
    test_clp_kernel = np.ones((2, n_in // 2, 128))
    network.initialize_params(session)
    test_clp_output = session.run(
      layer.output.placeholder,
      feed_dict={network.get_extern_data('data').placeholder: test_input, layer._clp_kernel: test_clp_kernel})
    assert test_clp_output[0, 0, 0] - 6.00722122 < 1e-5


def test_MultichannelStftLayer():
  def _get_ref_output(time_sig, fft_size, frame_size, frame_shift, window_name, frame_nr, channel_nr):
      import numpy as np
      frame_start = frame_nr * frame_shift
      frame_end = frame_start + frame_size
      frame = time_sig[0, frame_start:frame_end, channel_nr]
      if window_name == "hanning":
          window = np.hanning(frame_size)
      windowed_frame = window * frame
      out = np.fft.rfft(windowed_frame, fft_size)
      return out

  def test_rfftStftConfig_01():
    with make_scope() as session:
      layer_name = "stft_layer"
      fft_size = 400
      frame_size = 400
      frame_shift = 160
      window = "hanning"
      test_input = np.ones((1, 32000, 2), dtype=np.float32)
      config = Config()
      config.update({
        "num_outputs": int(fft_size / 2) + 1 * test_input.shape[2],
        "num_inputs": test_input.shape[2],
        "network": {
          layer_name: {
            "class": "multichannel_stft_layer",
            "frame_shift": frame_shift, "frame_size": frame_size, "window": window,
            "fft_size": fft_size, "use_rfft": True, "nr_of_channels": 2, "is_output_layer": True,
            "from": "data:data"}
        }})
      network = TFNetwork(config=config, train_flag=True)
      network.construct_from_dict(config.typed_value("network"))
      layer = network.layers[layer_name]
      test_output = session.run(layer.output.placeholder, {network.get_extern_data('data').placeholder: test_input})
      ref0 = _get_ref_output(test_input, fft_size, frame_size, frame_shift, window, 0, 0)
      # np.fft.rfft and tensorflow.python.ops.rfft differ a little bit in their
      # results, thus an error margin is allowed in the result
      resultDiff = np.abs(test_output[0, 0, 0:(int(fft_size / 2) + 1)] - ref0)
      assert np.mean(resultDiff) < 0.02
      assert np.max(resultDiff) < 1
      pass

  test_rfftStftConfig_01()


def test_MultichannelMultiResStftLayer():
  def _get_ref_output_single_res(time_sig, fft_size, frame_size, frame_shift, window_name, frame_nr, channel_nr):
      import numpy as np
      frame_start = frame_nr * frame_shift
      frame_end = frame_start + frame_size
      frame = time_sig[0, frame_start:frame_end, channel_nr]
      if window_name == "hanning":
          window = np.hanning(frame_size)
      windowed_frame = window * frame
      out = np.fft.rfft(windowed_frame, fft_size)
      return out

  def test_stftConfig_single_res_01():
    with make_scope() as session:
      layer_name = "stft_layer"
      fft_sizes = [400]
      frame_sizes = [400]
      frame_shift = 160
      window = "hanning"
      test_input = np.ones((1, 32000, 2), dtype=np.float32)
      num_outputs = (int(fft_sizes[0] / 2) + 1) * test_input.shape[2]
      config = Config()
      config.update({
        "num_outputs": num_outputs,
        "num_inputs": test_input.shape[2],
        "network": {
          layer_name: {
            "class": "multichannel_multiresolution_stft_layer",
            "frame_shift": frame_shift, "frame_sizes": frame_sizes, "window": window,
            "fft_sizes": fft_sizes, "use_rfft": True, "nr_of_channels": 2, "is_output_layer": True,
            "from": "data:data"}
        }})
      network = TFNetwork(config=config, train_flag=True)
      network.construct_from_dict(config.typed_value("network"))
      layer = network.layers[layer_name]
      test_output = session.run(layer.output.placeholder, {network.get_extern_data('data').placeholder: test_input})
      ref0 = _get_ref_output_single_res(test_input, fft_sizes[0], frame_sizes[0], frame_shift, window, 0, 0)
      resultDiff = np.abs(test_output[0, 0, 0:(int(fft_sizes[0] / 2) + 1)] - ref0)
      assert test_output.shape[2] == num_outputs
      assert np.mean(resultDiff) < 0.02
      assert np.max(resultDiff) < 1

  def test_stftConfig_multi_res_01():
    with make_scope() as session:
      layer_name = "stft_layer"
      fft_sizes = [400, 200]
      frame_sizes = [400, 200]
      frame_shift = 160
      window = "hanning"
      test_input = np.ones((1, 32000, 2), dtype=np.float32)
      test_input[0, 1000:1100, 1] = np.ones((100), dtype=np.float32) * 0.5
      num_outputs = int(np.sum([(int(fft_size / 2) + 1) * test_input.shape[2] for fft_size in fft_sizes]))
      config = Config()
      config.update({
        "num_outputs": num_outputs,
        "num_inputs": test_input.shape[2],
        "network": {
          layer_name: {
            "class": "multichannel_multiresolution_stft_layer",
            "frame_shift": frame_shift, "frame_sizes": frame_sizes, "window": window,
            "fft_sizes": fft_sizes, "use_rfft": True, "nr_of_channels": 2, "is_output_layer": True,
            "from": "data:data"}
        }})
      network = TFNetwork(config=config, train_flag=True)
      network.construct_from_dict(config.typed_value("network"))
      layer = network.layers[layer_name]
      test_output = session.run(layer.output.placeholder, {network.get_extern_data('data').placeholder: test_input})
      assert test_output.shape[2] == num_outputs
      comparison_frame = 6
      ref00 = _get_ref_output_single_res(test_input, fft_sizes[0], frame_sizes[0], frame_shift, window, comparison_frame, 0)
      ref01 = _get_ref_output_single_res(test_input, fft_sizes[0], frame_sizes[0], frame_shift, window, comparison_frame, 1)
      ref10 = _get_ref_output_single_res(test_input, fft_sizes[1], frame_sizes[1], frame_shift, window, comparison_frame, 0)
      ref11 = _get_ref_output_single_res(test_input, fft_sizes[1], frame_sizes[1], frame_shift, window, comparison_frame, 1)
      ref = np.concatenate([ref00, ref01, ref10, ref11], axis=0)
      resultDiff = np.abs(test_output[0, comparison_frame, :] - ref)
      assert np.mean(resultDiff) < 0.02
      assert np.max(resultDiff) < 1

  def test_stftConfig_multi_res_02():
    with make_scope() as session:
      layer_name = "stft_layer"
      fft_sizes = [400, 200, 800]
      frame_sizes = [400, 200, 800]
      frame_shift = 160
      window = "hanning"
      test_input = np.random.normal(0, 0.6, (1, 3200, 2))
      num_outputs = int(np.sum([(int(fft_size / 2) + 1) * test_input.shape[2] for fft_size in fft_sizes]))
      config = Config()
      config.update({
        "num_outputs": num_outputs,
        "num_inputs": test_input.shape[2],
        "network": {
          layer_name: {
            "class": "multichannel_multiresolution_stft_layer",
            "frame_shift": frame_shift, "frame_sizes": frame_sizes, "window": window,
            "fft_sizes": fft_sizes, "use_rfft": True, "nr_of_channels": 2, "is_output_layer": True,
            "from": "data:data"}
        }})
      network = TFNetwork(config=config, train_flag=True)
      network.construct_from_dict(config.typed_value("network"))
      layer = network.layers[layer_name]
      test_output = session.run(layer.output.placeholder, {network.get_extern_data('data').placeholder: test_input})
      assert test_output.shape[2] == num_outputs
      comparison_frame = 6
      ref00 = _get_ref_output_single_res(test_input, fft_sizes[0], frame_sizes[0], frame_shift, window, comparison_frame, 0)
      ref01 = _get_ref_output_single_res(test_input, fft_sizes[0], frame_sizes[0], frame_shift, window, comparison_frame, 1)
      ref10 = _get_ref_output_single_res(test_input, fft_sizes[1], frame_sizes[1], frame_shift, window, comparison_frame, 0)
      ref11 = _get_ref_output_single_res(test_input, fft_sizes[1], frame_sizes[1], frame_shift, window, comparison_frame, 1)
      ref20 = _get_ref_output_single_res(test_input, fft_sizes[2], frame_sizes[2], frame_shift, window, comparison_frame, 0)
      ref21 = _get_ref_output_single_res(test_input, fft_sizes[2], frame_sizes[2], frame_shift, window, comparison_frame, 1)
      ref = np.concatenate([ref00, ref01, ref10, ref11, ref20, ref21], axis=0)
      resultDiff = np.abs(test_output[0, comparison_frame, :] - ref)
      assert np.mean(resultDiff) < 0.06
      assert np.max(resultDiff) < 1

  test_stftConfig_single_res_01()
  test_stftConfig_multi_res_01()
  test_stftConfig_multi_res_02()


if __name__ == "__main__":
  better_exchook.install()
  if len(sys.argv) <= 1:
    for k, v in sorted(globals().items()):
      if k.startswith("test_"):
        print("-" * 40)
        print("Executing: %s" % k)
        try:
          v()
        except unittest.SkipTest as exc:
          print("SkipTest:", exc)
        print("-" * 40)
    print("Finished all tests.")
  else:
    assert len(sys.argv) >= 2
    for arg in sys.argv[1:]:
      print("Executing: %s" % arg)
      if arg in globals():
        globals()[arg]()  # assume function and execute
      else:
        eval(arg)  # assume Python code and execute




"""
Provides :class:`NumpyDumpDataset`.
"""

from returnn.datasets.basic import Dataset, DatasetSeq
import os
import numpy
import typing


class NumpyDumpDataset(Dataset):
  """
  For ``tools/dump-dataset.py --type=numpy``.
  """

  file_format_data = "%i.data"
  file_format_targets = "%i.targets"

  def __init__(self, prefix, postfix=".txt.gz",
               start_seq=0, end_seq=None,
               num_inputs=None, num_outputs=None, **kwargs):
    super(NumpyDumpDataset, self).__init__(**kwargs)
    self.file_format_data = prefix + self.file_format_data + postfix
    self.file_format_targets = prefix + self.file_format_targets + postfix
    self.start_seq = start_seq
    self._init_num_seqs(end_seq)
    self._seq_index = None
    self.cached_seqs = []  # type: typing.List[DatasetSeq]
    self.num_inputs = num_inputs
    self.num_outputs = num_outputs
    assert num_inputs and num_outputs

  def _init_num_seqs(self, end_seq=None):
    last_seq = None
    i = self.start_seq
    while True:
      if end_seq is not None and i >= end_seq:
        break
      if not os.path.exists(self.file_format_data % i):
        break
      if not os.path.exists(self.file_format_targets % i):
        break
      last_seq = i
      i += 1
    if end_seq is None:
      assert last_seq is not None, "None found. Check %s." % (self.file_format_data % self.start_seq)
      end_seq = last_seq
    else:
      assert last_seq == end_seq - 1, "Check %s." % (self.file_format_data % end_seq)
    assert end_seq > self.start_seq
    self._num_seqs = end_seq - self.start_seq

  def _load_numpy_seq(self, seq_idx):
    """
    :param int seq_idx:
    """
    real_idx = self._seq_index[seq_idx]
    features = numpy.loadtxt(self.file_format_data % real_idx)
    targets = numpy.loadtxt(self.file_format_targets % real_idx)
    assert features.ndim == 2
    assert features.shape[1] == self.num_inputs
    assert targets.ndim == 1
    self._add_cache_seq(seq_idx, features, targets)

  # ------------ Dataset API --------------

  def init_seq_order(self, epoch=None, seq_list=None, seq_order=None):
    """
    :param int|None epoch:
    :param list[str]|None seq_list:
    :param list[int]|None seq_order:
    :rtype: bool
    """
    super(NumpyDumpDataset, self).init_seq_order(epoch=epoch, seq_list=seq_list, seq_order=seq_order)
    if seq_list is not None or seq_order is not None:
      raise NotImplementedError
    if self.seq_ordering == "sorted":  # not supported atm
      self.seq_ordering = "default"
    self._seq_index = [i + self.start_seq for i in self.get_seq_order_for_epoch(epoch, self.num_seqs)]
    self.cached_seqs[:] = []
    return True

  def _load_seqs(self, start, end):
    """
    :param int start:
    :param int end:
    """
    self._cleanup_old_seq_cache(start)
    for i in range(start, end):
      if not self._have_cache_seq(i):
        self._load_numpy_seq(i)

  def get_input_data(self, seq_idx):
    """
    :param int seq_idx:
    :rtype: numpy.ndarray
    """
    return self._get_cache_seq(seq_idx).get_data("data")

  def get_targets(self, target, seq_idx):
    """
    :param str target:
    :param int seq_idx:
    :rtype: numpy.ndarray
    """
    return self._get_cache_seq(seq_idx).get_data(target)

  def get_ctc_targets(self, seq_idx):
    """
    Not implemented.

    :param int seq_idx:
    """
    assert False, "No CTC targets."

  def get_seq_length(self, seq_idx):
    """
    :param int seq_idx:
    :rtype: Util.NumbersDict
    """
    # This is different from the other get_* functions.
    # load_seqs() might not have been called before.
    if not self._have_cache_seq(seq_idx):
      self._load_numpy_seq(seq_idx)
    return self._get_cache_seq(seq_idx).num_frames

  @property
  def num_seqs(self):
    """
    :rtype: int
    """
    return self._num_seqs

  def len_info(self):
    """
    :rtype: str
    """
    return "%s, %i seqs" % (self.__class__.__name__, self.num_seqs)

  # ------------ Seq cache management -----------

  def _cleanup_old_seq_cache(self, seq_end):
    i = 0
    while i < len(self.cached_seqs):
      if self.cached_seqs[i].seq_idx >= seq_end:
        break
      i += 1
    del self.cached_seqs[:i]

  def _get_cache_seq(self, seq_idx, error_not_found=True):
    for data in self.cached_seqs:
      if data.seq_idx == seq_idx:
        return data
    if error_not_found:
      raise Exception("seq %i not loaded" % seq_idx)
    else:
      return None

  def _have_cache_seq(self, seq_idx):
    return self._get_cache_seq(seq_idx, error_not_found=False) is not None

  def _get_cache_last_seq_idx(self):
    if self.cached_seqs:
      return self.cached_seqs[-1].seq_idx
    else:
      return -1

  def _add_cache_seq(self, seq_idx, features, targets):
    last_seq_idx = self._get_cache_last_seq_idx()
    assert seq_idx == last_seq_idx + 1
    self.cached_seqs += [DatasetSeq(seq_idx, features, targets)]




from __future__ import print_function
import gc
import sys
import time
import numpy
import threading
import typing
from .basic import Dataset
from returnn.log import log
from returnn.util import NumbersDict


class CachedDataset(Dataset):

  def __init__(self, cache_byte_size=0, **kwargs):
    """
    :param int cache_byte_size:
    """
    super(CachedDataset, self).__init__(**kwargs)
    self.cache_byte_size_total_limit = cache_byte_size
    if cache_byte_size == -1:
      self.cache_byte_size_limit_at_start = 1024 ** 4
    elif cache_byte_size == 0:
      self.cache_byte_size_limit_at_start = 0
    else:
     self.cache_byte_size_limit_at_start = max(cache_byte_size * 2 // 3, 1)
     self.cache_byte_size_total_limit = max(cache_byte_size - self.cache_byte_size_limit_at_start, 1)
    self.num_seqs_cached_at_start = 0
    self.cached_bytes_at_start = 0
    self.nbytes = 0
    self.start_cache_initialized = False
    self.definite_cache_leftover = 0
    self.cache_num_frames_free = 0
    self.preload_set = set([])
    self.preload_end = 0
    self.max_ctc_length = 0
    self.ctc_targets = None
    self.alloc_intervals = None  # type: typing.Optional[list]
    self._seq_start = []  # [numpy.array([0,0])]  # uses sorted seq idx, see set_batching()
    self._seq_index = []  # type: typing.List[int]  # Via init_seq_order(). seq_index idx -> hdf seq idx
    self._seq_index_inv = {}  # type: typing.Dict[int,int]  # Via init_seq_order(). hdf seq idx -> seq_index idx
    self._index_map = range(len(self._seq_index))  # sorted seq idx -> seq_index idx
    self._tag_idx = {}  # type: typing.Dict[str,int]  # map of tag -> real-seq-idx. call _update_tag_idx
    self.targets = {}
    self.target_keys = []  # the keys for which we provide data; we may have labels for additional keys in self.labels
    self.timestamps = None

  def initialize(self):
    """
    Initialization.
    """
    super(CachedDataset, self).initialize()

    if self.cache_byte_size_limit_at_start > 0:
      self.nbytes = numpy.array([], dtype=numpy.float32).itemsize * (self.num_inputs * self.window + 1 + 1)

      # Calculate cache sizes.
      temp_cache_size_bytes = max(0, self.cache_byte_size_total_limit)
      self.definite_cache_leftover = temp_cache_size_bytes if self.num_seqs_cached_at_start == self.num_seqs else 0
      self.cache_num_frames_free = temp_cache_size_bytes // self.nbytes

      print("cached %i seqs" % self.num_seqs_cached_at_start,
            "%s GB" % (self.cached_bytes_at_start / float(1024 * 1024 * 1024)),
            ("(fully loaded, %s GB left over)" if self.definite_cache_leftover else "(%s GB free)") %
            max(temp_cache_size_bytes / float(1024 * 1024 * 1024), 0),
            file=log.v4)

  def init_seq_order(self, epoch=None, seq_list=None, seq_order=None):
    """
    :type epoch: int|None
    :param list[str]|None seq_list: List of sequence tags, to set a predefined order.
    :param list[int]|None seq_order: List of corpus sequence indices, to set a predefined order.
    Initialize lists:
      self.seq_index  # sorted seq idx
    """
    super(CachedDataset, self).init_seq_order(epoch=epoch, seq_list=seq_list, seq_order=seq_order)
    if seq_order is not None:
      seq_index = seq_order
    elif seq_list is not None:
      self._update_tag_idx()
      seq_index = [self._tag_idx[tag] for tag in seq_list]
    else:
      seq_index = self.get_seq_order_for_epoch(epoch, self._num_seqs, lambda s: self._get_seq_length_by_real_idx(s)[0])

    old_index_map = self._index_map[:]
    self._index_map = range(len(seq_index))  # sorted seq idx -> seq_index idx

    if isinstance(seq_index, numpy.ndarray) or isinstance(self._seq_index, numpy.ndarray):
      seq_index_unchanged = numpy.array_equal(self._seq_index, seq_index)
    else:
      seq_index_unchanged = (self._seq_index == seq_index)

    if seq_index_unchanged and self.start_cache_initialized:
      return False

    if epoch is not None:
      # Give some hint to the user in case he is wondering why the cache is reloading.
      print("Reinitialize dataset seq order for epoch %i." % epoch, file=log.v4)

    if (self.cache_byte_size_limit_at_start == 0
        or self.num_seqs_cached_at_start != len(seq_index)
        or not self.start_cache_initialized):
      self._seq_index = seq_index
      self._seq_index_inv = {}  # reset, create later if needed
      self._init_seq_starts()
      self._init_alloc_intervals()
      self._init_start_cache()
      self.start_cache_initialized = True
    else:
      if not self._seq_index_inv:
        self._seq_index_inv = dict(zip(self._seq_index, range(len(self._seq_index))))  # hdf seq idx -> seq_index idx
      self._index_map = [self._seq_index_inv[i] for i in seq_index]  # sorted seq idx -> seq_index idx
      if self._index_map == old_index_map:
        return False
    return True

  def get_current_seq_order(self):
    assert self.cache_byte_size_limit_at_start == 0  # not implemented otherwise, we ignore _index_map
    return self._seq_index

  def _get_tag_by_real_idx(self, real_idx):
    raise NotImplementedError

  def _update_tag_idx(self):
    if self._tag_idx:
      return
    for i in range(self._num_seqs):
      self._tag_idx[self._get_tag_by_real_idx(i)] = i

  def batch_set_generator_cache_whole_epoch(self):
    return True

  def _init_alloc_intervals(self):
    if self.cache_byte_size_limit_at_start == 0:
      return
    assert self.num_seqs > 0
    assert self.num_inputs > 0
    assert self.window > 0
    self.preload_set = set([])
    self.alloc_intervals = \
      [(0, 0, numpy.zeros([1] + self.get_data_shape("data"), dtype=self.get_data_dtype("data"))),
       (self.num_seqs, self.num_seqs, numpy.zeros([1] + self.get_data_shape("data"), dtype=self.get_data_dtype("data")))]
    # self.alloc_intervals[i] is (idx start, idx end, data), where
    # idx start/end is the sorted seq idx start/end, end exclusive,
    # and data is a numpy.array.

  def _init_seq_starts(self):
    if self.cache_byte_size_limit_at_start == 0:
      return
    self._seq_start = [self._seq_start[0] * 0]  # idx like in seq_index, *not* real idx
    for i in range(self.num_seqs):
      ids = self._seq_index[i]
      self._seq_start.append(self._seq_start[-1] + self._get_seq_length_by_real_idx(ids))

  def _init_start_cache(self):
    if self.cache_byte_size_limit_at_start == 0:
      return
    if not self.alloc_intervals:
      return
    if not self.nbytes:
      return

    num_cached = 0
    cached_bytes = 0
    for i in range(self.num_seqs):
      if i == num_cached:
        nbytes = self.get_seq_length_nd(i)[0] * self.nbytes
        if self.cache_byte_size_limit_at_start >= cached_bytes + nbytes:
          num_cached = i + 1
          cached_bytes += nbytes

    self.num_seqs_cached_at_start = num_cached
    self.cached_bytes_at_start = cached_bytes
    if num_cached > 0:
      self.preload_end = num_cached
      if sys.version_info >= (3, 0):
        threading.Thread(target=self._preload_seqs, args=(0, num_cached), daemon=True).start()
      else:
        threading.Thread(target=self._preload_seqs, args=(0, num_cached)).start()

  def load_seqs(self, start, end):
    """
    Load data sequences.
    As a side effect, will modify / fill-up:
      self.alloc_intervals
      self.targets
    This does some extra logic for the cache and calls self._load_seqs()
    for the real loading.

    :param int start: start sorted seq idx
    :param int end: end sorted seq idx
    """
    assert start >= 0
    assert start <= end

    if self.is_cached(start, end, blocking=True):
      return

    if self.cache_byte_size_limit_at_start > 0:  # If the cache is enabled.
      self._load_seqs_with_cache(start, end)
      return self.is_cached(start, end, blocking=True)

    super(CachedDataset, self).load_seqs(start, end)

  def _load_seqs(self, start, end):
    raise NotImplementedError

  def _load_seqs_with_cache(self, start, end, clear=True):
    if not clear:
      # only remove as many frames as required
      num_needed_cache_frames = self.get_seq_start(end)[0] - self.get_seq_start(start)[0]
      if self.cache_num_frames_free < num_needed_cache_frames:
        self.cache_num_frames_free += self.delete(num_needed_cache_frames - self.cache_num_frames_free)
        gc.collect()
      self.cache_num_frames_free -= num_needed_cache_frames
      threading.Thread(target=self._preload_seqs,args=(start,end)).start()
    else:
      # First, delete everything.
      self.cache_num_frames_free += self.delete(None)
      gc.collect()
      # Preload as much as we can so that we fill up the cache.
      while end < self.num_seqs:
        num_needed_cache_frames = self.get_seq_length_nd(end)[0]
        if self.cache_num_frames_free - num_needed_cache_frames < 0:
          break
        self.cache_num_frames_free -= num_needed_cache_frames
        end += 1
      self.preload_end = end
      threading.Thread(target=self._preload_seqs,args=(start,end)).start()

  def _preload_seqs(self,start,end):
    print("Preloading cache from", start, "to", end, file=log.v4)
    super(CachedDataset, self).load_seqs(start, end)
    self.preload_end = self.num_seqs_cached_at_start

  def _shuffle_frames_in_seqs(self, start, end):
    """
    :type start: int
    :type end: int
    """
    assert start < end
    assert self.is_cached(start, end)
    alloc_idx = self.alloc_interval_index(start)
    alloc_start, alloc_end, alloc_data = self.alloc_intervals[alloc_idx]
    assert start >= alloc_start
    assert end <= alloc_end
    rnd = numpy.random.RandomState(start)  # Some deterministic way to shuffle!
    num_frames = self._seq_start[end][0] - self._seq_start[start][0]
    assert num_frames > 0
    perm = rnd.permutation(num_frames)
    alloc_offset = self._seq_start[start][0] - self._seq_start[alloc_start][0]
    assert alloc_offset + num_frames <= alloc_data.shape[0]
    # Permute alloc_data.
    data = alloc_data[alloc_offset:alloc_offset + num_frames]
    alloc_data[alloc_offset:alloc_offset + num_frames] = data[perm]
    # Permute targets.
    for k in self.targets:
      idx = self.target_keys.index(k) + 1
      targets = self.targets[k][self._seq_start[idx]:self._seq_start[start][idx] + num_frames]
      self.targets[k][self._seq_start[start][idx]:self._seq_start[start][idx] + self._seq_start[end][idx] - self._seq_start[start][idx]] = targets[perm]

  def _set_alloc_intervals_data(self, idc, data):
    """
    :param int idc: index of sorted seq idx
    :param numpy.ndarray data: raw data
    """
    idi = self.alloc_interval_index(idc)
    assert idi >= 0
    o = self._seq_start[idc][0] - self._seq_start[self.alloc_intervals[idi][0]][0]
    l = data.shape[0]
    x = data
    if self.window > 1:
      x = self._sliding_window(x)
    self.alloc_intervals[idi][2][o:o + l] = x

  def alloc_interval_index(self, ids):
    """
    :param int ids: sorted seq idx
    :return index in self.alloc_intervals
    :rtype: int
    """
    s = 0
    e = len(self.alloc_intervals)
    # Binary search.
    while s < e:
      i = (s + e) // 2
      alloc_start, alloc_end, _ = self.alloc_intervals[i]
      if alloc_start <= ids < alloc_end:
        return i
      elif alloc_start <= ids and ids >= alloc_end:
        if s == i:
          return -1
        s = i
      elif alloc_start > ids:
        if e == i:
          return -1
        e = i
      else:
        assert False
    return -1

  def _insert_alloc_interval(self, pos, value, merge=False):
    """
    Insert np.zeros into self.alloc_intervals.
    :param int pos: idx in self.alloc_intervals
    :param (int,int) value: (start,end) like in load_seqs(), sorted seq idx
    :rtype: int
    """
    if value[0] == value[1]:
      return 0
    ci = self.alloc_intervals[pos][1]
    ni = self.alloc_intervals[pos + 1][0]
    xc = self.alloc_intervals[pos][2]
    xn = self.alloc_intervals[pos + 1][2]
    if value[0] == ci and value[1] == ni and merge:
      nj = self.alloc_intervals[pos][0]
      nk = self.alloc_intervals[pos + 1][1]
      del self.alloc_intervals[pos]
      del self.alloc_intervals[pos]
      self.alloc_intervals.insert(pos,
        (nj,nk,
         numpy.concatenate(
           [xc,
            numpy.zeros(
              [self._seq_start[ni][0]] + self.get_data_shape("data"),
              dtype=self.get_data_dtype("data")),
            xn])))
      return 0
    elif value[0] == ci and merge:
      nj = self.alloc_intervals[pos][0]
      del self.alloc_intervals[pos]
      self.alloc_intervals.insert(pos, (nj,value[1],
                                        numpy.concatenate([xc, numpy.zeros([self._seq_start[value[1]][0] - self._seq_start[ci][0]] + self.get_data_shape("data"), dtype=self.get_data_dtype("data"))])))
      return 0
    elif value[1] == ni and merge:
      nk = self.alloc_intervals[pos + 1][1]
      del self.alloc_intervals[pos + 1]
      self.alloc_intervals.insert(pos + 1, (value[0], nk,
                                            numpy.concatenate([numpy.zeros([self._seq_start[ni][0] - self._seq_start[value[0]][0]] + self.get_data_shape("data"), dtype=self.get_data_dtype("data")), xc])))
      return 0
    else:
      self.alloc_intervals.insert(pos + 1,
        value + (numpy.zeros(
            [self._seq_start[value[1]][0] - self._seq_start[value[0]][0]] + self.get_data_shape("data"),
            dtype=self.get_data_dtype("data")),))
      return 1

  def _remove_alloc_interval(self, pos, value):
    """
    Remove data from self.alloc_intervals.
    :param int pos: idx in self.alloc_intervals
    :param (int,int) value: (start,end) like in load_seqs(), sorted seq idx
    :rtype: int
    """
    ci, ni, xi = self.alloc_intervals[pos]
    if value[0] == ci and value[1] == ni:
      del self.alloc_intervals[pos]
      return -1
    elif value[0] == ci:
      self.alloc_intervals.insert(pos, (value[1], ni, xi[self._seq_start[value[1]][0] - self._seq_start[ci][0]:]))
      del self.alloc_intervals[pos + 1]
      return 0
    elif value[1] == ni:
      self.alloc_intervals.insert(pos, (ci, value[0], xi[:self._seq_start[value[0]][0] - self._seq_start[ci][0]]))
      del self.alloc_intervals[pos + 1]
      return 0
    else:
      self.alloc_intervals.insert(pos, (value[1], ni, xi[self._seq_start[value[1]][0] - self._seq_start[ci][0]:]))
      self.alloc_intervals.insert(pos, (ci, value[0], xi[:self._seq_start[value[0]][0] - self._seq_start[ci][0]]))
      del self.alloc_intervals[pos + 2]
      return 1

  def _modify_alloc_intervals(self, start, end, invert):
    """
    Inserts/removes sorted seq idx range (start,end).
    :param int start: like in load_seqs(), sorted seq idx
    :param int end: like in load_seqs(), sorted seq idx
    :param bool invert: True->insert, False->remove
    :rtype: list[int]
    :return selection list, modified sorted seq idx in self.alloc_intervals
    """
    if end is None:
      end = start + 1
    if start == end:
      return
    assert start < end
    i = 0
    selection = []  # type: typing.List[int]
    modify = self._insert_alloc_interval if invert else self._remove_alloc_interval
    while i < len(self.alloc_intervals) - invert:
      ni = self.alloc_intervals[i + invert][1 - invert]  # insert mode: start idx of next alloc
      ci = self.alloc_intervals[i][invert]               # insert mode: end idx of cur alloc
      assert ci <= ni
      flag = ((ci <= start < ni), (ci < end <= ni), (ci < start and ni <= start) or (ci >= end and ni > end))
      if not flag[0] and not flag[1]:
        if not flag[2]:
          selection.extend(range(ci, ni))
          i += modify(i, (ci, ni))
      elif flag[1]:
        v = (start if flag[0] else ci, end)
        selection.extend(range(v[0], v[1]))
        i += modify(i, v)
        break
      elif flag[0]:
        selection.extend(range(start, ni))
        i += modify(i, (start, ni))
      i += 1
    if self.alloc_intervals[0][0] != 0:
      self.alloc_intervals.insert(0, (0, 0, numpy.zeros([1] + self.get_data_shape("data"), dtype=self.get_data_dtype("data"))))
    if self.alloc_intervals[-1][1] != self.num_seqs:
      self.alloc_intervals.append((self.num_seqs, self.num_seqs, numpy.zeros([1] + self.get_data_shape("data"), dtype=self.get_data_dtype("data"))))
    return selection

  def insert_alloc_interval(self, start, end=None):
    return self._modify_alloc_intervals(start, end, True)

  def remove_alloc_interval(self, start, end=None):
    return self._modify_alloc_intervals(start, end, False)

  def delete(self, nframes):
    """
    :param int|None nframes: how much frames to delete max.
      Note that this limit is not strict. We can end up
      deleting more than nframes.
    :return: number of frames deleted
    :rtype: int
    """
    if nframes is not None:
      if nframes == 0:
        return 0
      assert nframes > 0
    deleted = 0
    i = 0
    while (not nframes or deleted < nframes) and i < len(self.alloc_intervals):
      ai = self.alloc_intervals[i]
      if ai[1] > self.num_seqs_cached_at_start and ai[0] < ai[1]:
        removed = self.remove_alloc_interval(max(ai[0],self.num_seqs_cached_at_start), ai[1])
        self.preload_set -= set(removed)
        deleted += sum([self._get_seq_length_by_real_idx(self._seq_index[i])[0] for i in removed])
      else:
        i += 1
    return deleted

  @property
  def num_seqs(self):
    return len(self._index_map)

  def is_cached(self, start, end, blocking = False):
    """
    :param int start: like in load_seqs(), sorted seq idx
    :param int end: like in load_seqs(), sorted seq idx
    :rtype: bool
    :returns whether we have the full range (start,end) of sorted seq idx
      cached in self.alloc_intervals (end is exclusive).
    """
    if self.cache_byte_size_total_limit == 0:  # disabled cache
      return False
    if start == end:
      return True  # Empty.
    assert start < end
    if blocking and end <= self.preload_end:
      while not set(range(start,end)) <= self.preload_set:
        time.sleep(0.2)
      return True
    return set(range(start,end)) <= self.preload_set

  def _get_seq_length_by_real_idx(self, real_seq_idx):
    """
    :param int real_seq_idx:
    :returns length of the sequence with index 'real_seq_idx'
    :rtype: numpy.ndarray
    """
    raise NotImplementedError

  def get_seq_length_nd(self, sorted_seq_idx):
    """
    :type sorted_seq_idx: int
    :rtype: numpy.ndarray
    """
    real_seq_idx = self._seq_index[self._index_map[sorted_seq_idx]]
    return self._get_seq_length_by_real_idx(real_seq_idx)

  def get_seq_length(self, seq_idx):
    """
    :rtype: NumbersDict
    """
    lengths = self.get_seq_length_nd(seq_idx)
    d = {}
    first_target_idx = 0
    # We allow using only targets. In this case self.num_inputs == 0 and the "data" key is not used.
    if self.num_inputs > 0:
      d["data"] = lengths[0]
      first_target_idx = 1
    for k, l in zip(self.target_keys, lengths[first_target_idx:]):
      d[k] = l
    return NumbersDict(d)

  def get_seq_start(self, sorted_seq_idx):
    """
    :type sorted_seq_idx: int
    :rtype: (int,int)
    """
    return self._seq_start[sorted_seq_idx]

  def get_times(self, sorted_seq_idx):
    seq_start = self.get_seq_start(sorted_seq_idx)[0]
    seq_len = self.get_seq_length_nd(sorted_seq_idx)[0]
    return self.timestamps[seq_start:seq_start + seq_len]

  def get_input_data(self, sorted_seq_idx):
    seq_idx = self._index_map[sorted_seq_idx]
    idi = self.alloc_interval_index(seq_idx)
    assert idi >= 0, "failed to get data for seq %i" % sorted_seq_idx
    alloc_start_seq, alloc_end_seq, alloc_data = self.alloc_intervals[idi]
    o = self.get_seq_start(seq_idx)[0] - self.get_seq_start(alloc_start_seq)[0]
    assert o >= 0
    l = self.get_seq_length_nd(sorted_seq_idx)[0]
    assert alloc_data.shape[0] >= o + l
    return alloc_data[o:o + l]

  def get_data_dim(self, key):
    if key == "data" and self.num_inputs > 0:  # if num_inputs == 0, we allow "data" as a target key
      return self.num_inputs * self.window
    return self.num_outputs[key][0]

  def get_targets(self, target, sorted_seq_idx):
    seq_idx = self._index_map[sorted_seq_idx]
    idx = self.target_keys.index(target) + 1
    seq_start = self.get_seq_start(seq_idx)[idx]
    seq_len = self.get_seq_length_nd(sorted_seq_idx)[idx]
    return self.targets[target][seq_start:seq_start + seq_len]

  def get_target_list(self):
    return list(self.targets.keys())

  def get_ctc_targets(self, sorted_seq_idx):
    ids = self._seq_index[self._index_map[sorted_seq_idx]]
    return self.ctc_targets[ids]

  def has_ctc_targets(self):
    return self.ctc_targets is not None

  def get_tag(self, sorted_seq_idx):
    raise NotImplementedError

  def have_corpus_seq_idx(self):
    return True

  def get_corpus_seq_idx(self, seq_idx):
    """
    :param int seq_idx: sorted sequence index from the current epoch, depending on seq_ordering
    :return: the sequence index as-is in the original corpus. only defined if self.have_corpus_seq_idx()
    :rtype: int
    """
    return self._seq_index[self._index_map[seq_idx]]




"""
Implements the SprintDatasetBase and ExternSprintDataset classes, some Dataset subtypes.
Note that from the main RETURNN process, you probably want ExternSprintDataset instead.
"""

from __future__ import print_function

import atexit
import os
import signal
import sys
try:
  import thread
except ImportError:
  import _thread as thread
from threading import Condition, currentThread, Thread
import time
import numpy
import typing

import returnn.util.task_system as task_system
from returnn.datasets.basic import Dataset, DatasetSeq
from .cached2 import CachedDataset2
from returnn.log import log
from returnn.util.task_system import Unpickler, numpy_copy_and_set_unused
from returnn.util.basic import eval_shell_str, interrupt_main, unicode, PY3, BytesIO, close_all_fds_except


class SprintDatasetBase(Dataset):
  """
  In Sprint, we use this object for multiple purposes:
  - Multiple epoch handling via SprintInterface.getSegmentList().
    For this, we get the segment list from Sprint and use the Dataset
    shuffling method.
  - Fill in data which we get via SprintInterface.feedInput*().
    Note that each such input doesn't necessarily correspond to a single
    segment. This depends which type of FeatureExtractor is used in Sprint.
    If we use the BufferedFeatureExtractor in utterance mode, we will get
    one call for every segment and we get also segmentName as parameter.
    Otherwise, we will get batches of fixed size - in that case,
    it doesn't correspond to the segments.
    In any case, we use this data as-is as a new seq.
    Because of that, we cannot really know the number of seqs in advance,
    nor the total number of time frames, etc.

  If you want to use this directly in RETURNN, see ExternSprintDataset.
  """

  SprintCachedSeqsMax = 200
  SprintCachedSeqsMin = 100

  def __init__(self, target_maps=None, str_add_final_zero=False, input_stddev=1.,
               orth_post_process=None, bpe=None, orth_vocab=None,
               suppress_load_seqs_print=False, reduce_target_factor=1,
               **kwargs):
    """
    :param dict[str,str|dict] target_maps: e.g. {"speaker_name": "speaker_map.txt"}, with "speaker_map.txt" containing
      a line for each expected speaker. The indices will be given by the line index. Note that scalar
      content (e.g. single index) will automatically get a time axis added with the length of the audio frames.
    :param bool str_add_final_zero: adds e.g. "orth0" with '\0'-ending
    :param float input_stddev: if != 1, will divide the input "data" by that
    :param str|list[str]|((str)->str)|None orth_post_process: :func:`get_post_processor_function`, applied on orth
    :param None|dict[str] bpe: if given, will be opts for :class:`BytePairEncoding`
    :param None|dict[str] orth_vocab: if given, orth_vocab is applied to orth and orth_classes is an available target`
    :param bool suppress_load_seqs_print: less verbose
    :param int reduce_target_factor: downsample factor to allow less targets than features
    """
    super(SprintDatasetBase, self).__init__(**kwargs)
    self.suppress_load_seqs_print = suppress_load_seqs_print
    self.reduce_target_factor = reduce_target_factor
    assert isinstance(self.reduce_target_factor, int) and self.reduce_target_factor >= 1
    if target_maps:
      assert isinstance(target_maps, dict)
      target_maps = target_maps.copy()
      for key, tmap in list(target_maps.items()):
        if isinstance(tmap, (str, unicode)):
          tmap = {l: i for (i, l) in enumerate(open(tmap).read().splitlines())}
        assert isinstance(tmap, dict)  # dict[str,int]
        target_maps[key] = tmap
    self.target_maps = target_maps
    self.str_add_final_zero = str_add_final_zero
    self.input_stddev = input_stddev
    # Note: "orth" is actually the raw bytes of the utf8 string,
    # so it does not make quite sense to associate a single str to each byte.
    # However, some other code might expect that the labels are all strings, not bytes,
    # and the API requires the labels to be strings.
    # The code in Dataset.serialize_data tries to decode this case as utf8 (if possible).
    self.labels["orth"] = [chr(i) for i in range(255)]
    self.orth_post_process = None  # type: typing.Optional[typing.Callable[[str],str]]
    if orth_post_process:
      if callable(orth_post_process):
        self.orth_post_process = orth_post_process
      else:
        from .lm import get_post_processor_function
        self.orth_post_process = get_post_processor_function(orth_post_process)
    self.bpe = None
    if bpe:
      from returnn.datasets.util.vocabulary import BytePairEncoding
      self.bpe = BytePairEncoding(**bpe)
      self.labels["bpe"] = self.bpe.labels
    self.orth_vocab = None
    if orth_vocab:
      from returnn.datasets.util.vocabulary import Vocabulary
      self.orth_vocab = Vocabulary.create_vocab(**orth_vocab)
      self.labels["orth_classes"] = self.orth_vocab.labels
    self.cond = Condition(lock=self.lock)
    self.add_data_thread_id = thread.get_ident()  # This will be created in the Sprint thread.
    self.ready_for_data = False
    self.reached_final_seq = False
    self.reached_final_seq_seen_all = False
    self.multiple_epochs = False
    self._complete_frac = None
    self.sprint_epoch = None  # in SprintInterface.getSegmentList()
    self.returnn_epoch = None  # in RETURNN train thread, Engine.train(). set via init_seq_order
    self.predefined_seq_list_order = None  # via init_seq_order
    self.sprint_finalized = False
    self._target_black_list = []  # if we get non numpy arrays and cannot convert them
    self._reset_cache()
    assert self.shuffle_frames_of_nseqs == 0  # Currently broken. But just use Sprint itself to do this.

  def use_multiple_epochs(self):
    """
    Called via SprintInterface.getSegmentList().
    """
    self.multiple_epochs = True

  def set_dimensions(self, input_dim, output_dim):
    """
    :type input_dim: int
    :type output_dim: int

    Called via python_train.
    """
    assert input_dim > 0
    self.num_inputs = input_dim
    self.num_outputs = {"data": (input_dim * self.window, 2)}
    if output_dim > 0:
      self.num_outputs["classes"] = (output_dim, 1)
    if self.bpe:
      self.num_outputs["bpe"] = (self.bpe.num_labels, 1)
    if self.orth_vocab:
      self.num_outputs["orth_classes"] = (self.orth_vocab.num_labels, 1)
    self.num_outputs["orth"] = (256, 1)
    if self.target_maps and "speaker_name" in self.target_maps:
      self.num_outputs["speaker_name"] = (len(self.target_maps["speaker_name"]), 1)
    else:
      self.num_outputs["speaker_name"] = (256, 1)
    self._base_init()
    # At this point, we are ready for data. In case we don't use the Sprint PythonSegmentOrdering
    # (SprintInterface.getSegmentList()), we must call this at least once.
    if not self.multiple_epochs:
      self.init_sprint_epoch(None)

  def _reset_cache(self):
    self.expected_load_seq_start = 0
    self.requested_load_seq_end = 0
    self.next_seq_to_be_added = 0
    self.reached_final_seq = False
    self.reached_final_seq_seen_all = False
    self._num_timesteps = 0
    self.added_data = []  # type: typing.List[DatasetSeq]
    self.ready_for_data = True

  def init_sprint_epoch(self, epoch):
    """
    :type epoch: int | None
    Called by SprintInterface.getSegmentList() when we start a new epoch.
    We must not call this via self.init_seq_order() because we will already have filled the cache by Sprint
    before the RETURNN train thread starts the epoch.
    """
    with self.lock:
      self.sprint_epoch = epoch
      self.sprint_finalized = False
      self._reset_cache()
      self.cond.notify_all()

  def finalize_sprint(self):
    """
    Called when SprintInterface.getSegmentList() ends.
    """
    with self.lock:
      self.sprint_finalized = True
      self.cond.notify_all()

  def init_seq_order(self, epoch=None, seq_list=None, seq_order=None):
    """
    Called by RETURNN train thread when we enter a new epoch.
    """
    if seq_order is not None:
      raise NotImplementedError("Predefined sequence order via indices in SprintDataset.")
    super(SprintDatasetBase, self).init_seq_order(epoch=epoch, seq_list=seq_list, seq_order=seq_order)
    if self.orth_vocab:
      self.orth_vocab.set_random_seed(self._get_random_seed_for_epoch(epoch=epoch))
    with self.lock:
      self.returnn_epoch = epoch
      self.predefined_seq_list_order = seq_list
      self.cond.notify_all()
      # No need to wait/check for Sprint thread here.
      # SprintInterface.getSegmentList() will wait for us.
    return True

  def _cleanup_old_seq_cache(self, seq_end):
    i = 0
    while i < len(self.added_data):
      if self.added_data[i].seq_idx >= seq_end:
        break
      i += 1
    del self.added_data[:i]

  def wait_for_returnn_epoch(self, epoch):
    """
    Called by SprintInterface.
    """
    with self.lock:
      while epoch != self.returnn_epoch:
        assert epoch > self.returnn_epoch
        self.cond.wait()

  def _wait_for_seq_can_pass_check(self, seq_start, seq_end):
    """
    :param int seq_start:
    :param int seq_end:
    :return: True if _waitForSeq can pass/return. False means that we need to wait more (until next signal)
    :rtype: bool
    """
    if self.reached_final_seq:
      return True
    if self._have_seqs_added(seq_start, seq_end):
      return True
    return False

  def _wait_for_seq(self, seq_start, seq_end=None):
    """
    Called by RETURNN train thread.
    Wait until we have seqs [seqStart,..,seqEnd-1] loaded,
    or we now that they will not be loaded anymore because we reached the end.

    :param int seq_start:
    :param int|None seq_end:
    """
    if seq_end is None:
      seq_end = seq_start + 1
    if seq_end > self.requested_load_seq_end:
      self.requested_load_seq_end = seq_end
      self.cond.notify_all()
    if self._wait_for_seq_can_pass_check(seq_start=seq_start, seq_end=seq_end):
      return
    # We need to wait.
    assert thread.get_ident() != self.add_data_thread_id
    if not self.suppress_load_seqs_print:
      print(
        "%s %s: wait for seqs (%i,%i) (last added: %s) (current time: %s)" % (
          self, currentThread().name, seq_start, seq_end, self._latest_added_seq(), time.strftime("%H:%M:%S")),
        file=log.v5)
    while not self._wait_for_seq_can_pass_check(seq_start=seq_start, seq_end=seq_end):
      self.cond.wait()

  def _latest_added_seq(self):
    if not self.added_data:
      return None
    return self.added_data[-1].seq_idx

  def _have_seqs_added(self, start, end=None):
    if end is None:
      end = start + 1
    if start >= end:
      return True
    for data in self.added_data:
      assert start >= data.seq_idx, "%s: We expect that we only ask about the cache of the upcoming seqs." % self
      if data.seq_idx == start:
        start += 1
      if start >= end:
        return True
    return False

  def _get_seq(self, seq_idx):
    """
    :param int seq_idx:
    :rtype: DatasetSeq
    """
    for data in self.added_data:
      if data.seq_idx == seq_idx:
        return data
    return None

  def is_cached(self, start, end):
    """
    :param int start:
    :param int end:
    :rtype: bool
    """
    # Always False, to force that we call self._load_seqs().
    # This is important for our buffer management.
    return False

  def load_seqs(self, start, end):
    """
    Called by RETURNN train thread.

    :param int start:
    :param int end:
    """
    if start == end:
      return
    if not self.suppress_load_seqs_print:
      print("%s load_seqs in %s:" % (self, currentThread().name), start, end, end=' ', file=log.v5)
    with self.lock:
      super(SprintDatasetBase, self).load_seqs(start, end)
      if not self.suppress_load_seqs_print:
        print("first features shape:", self._get_seq(start).features["data"].shape, file=log.v5)

  def _load_seqs(self, start, end):
    """
    Called by RETURNN train thread.
    We expect that start increase monotonic on each call
    for not-yet-loaded data.
    This will already be called with _load_seqs_superset indices.

    :param int start:
    :param int end:
    """
    assert start >= self.expected_load_seq_start
    if start > self.expected_load_seq_start:
      # Cleanup old data.
      self._cleanup_old_seq_cache(start)
      self.expected_load_seq_start = start
      self.cond.notify_all()
    self._wait_for_seq(start, end)

  def add_new_data(self, features, targets=None, segment_name=None):
    """
    Adds a new seq.
    This is called via the Sprint main thread.

    :param numpy.ndarray features: format (input-feature,time) (via Sprint)
    :param dict[str,numpy.ndarray|str] targets: format (time) (idx of output-feature)
    :param str|None segment_name:
    :returns the sorted seq index
    :rtype: int
    """

    # is in format (feature,time)
    assert self.num_inputs == features.shape[0]
    num_frames = features.shape[1]
    # features maybe downsampled via the network
    reduce_num_frames = int((num_frames + self.reduce_target_factor - 1) / self.reduce_target_factor)
    # must be in format: (time,feature)
    features = features.transpose()
    assert features.shape == (num_frames, self.num_inputs)
    if self.input_stddev != 1:
      features /= self.input_stddev
    if self.window > 1:
      features = self._sliding_window(features)
      assert features.shape == (num_frames, self.num_inputs * self.window)

    if targets is None:
      targets = {}
    if not isinstance(targets, dict):
      targets = {"classes": targets}
    if "classes" in targets:
      # 'classes' is always the alignment
      assert targets["classes"].shape == (reduce_num_frames,), (  # is in format (time,)
        "Number of targets %s does not match number of features %s (reduce factor %d)"
        % (targets["classes"].shape, (num_frames,), self.reduce_target_factor))
    if "speaker_name" in targets:
      targets["speaker_name"] = targets["speaker_name"].decode("utf8").strip()
    if "orth" in targets:
      targets["orth"] = targets["orth"].decode("utf8").strip()
    if "orth" in targets and self.orth_post_process:
      targets["orth"] = self.orth_post_process(targets["orth"])
    if self.bpe:
      assert "orth" in targets
      orth = targets["orth"]
      assert isinstance(orth, (str, unicode))
      assert "bpe" not in targets
      targets["bpe"] = numpy.array(self.bpe.get_seq(orth), dtype="int32")
    if self.orth_vocab:
      assert "orth" in targets
      orth = targets["orth"]
      assert isinstance(orth, (str, unicode))
      assert "orth_classes" not in targets
      targets["orth_classes"] = numpy.array(self.orth_vocab.get_seq(orth), dtype="int32")

    # Maybe convert some targets.
    if self.target_maps:
      for key, target_map in self.target_maps.items():
        assert key in targets
        v = target_map[targets[key]]
        v = numpy.asarray(v)
        if v.ndim == 0:
          v = numpy.zeros((num_frames,), dtype=v.dtype) + v  # add time dimension
        targets[key] = v

    # Maybe remove some targets.
    for key in self._target_black_list:
      if key in targets:
        del targets[key]

    # Check if all targets are valid.
    for key, v in sorted(list(targets.items())):
      if isinstance(v, numpy.ndarray):
        continue  # ok
      if isinstance(v, unicode):
        v = v.encode("utf8")
      if isinstance(v, (str, bytes)):
        if PY3:
          assert isinstance(v, bytes)
          v = list(v)
        else:
          v = list(map(ord, v))
        v = numpy.array(v, dtype="uint8")
        targets[key] = v
        if self.str_add_final_zero:
          v = numpy.append(v, numpy.array([0], dtype=v.dtype))
          assert key + "0" not in targets
          targets[key + "0"] = v
        continue
      print("%s, we will ignore the target %r because it is not a numpy array: %r" % (self, key, v), file=log.v3)
      self._target_black_list += [key]
      del targets[key]

    with self.lock:
      # This gets called in the Sprint main thread.
      # If this is used together with SprintInterface.getSegmentList(), we are always in a state where
      # we just yielded a segment name, thus we are always in a Sprint epoch and thus ready for data.
      assert self.ready_for_data
      assert not self.reached_final_seq
      assert not self.sprint_finalized

      seq_idx = self.next_seq_to_be_added

      if self.predefined_seq_list_order:
        # Note: Only in ExternSprintDataset, we can reliably set the seq order for now.
        assert seq_idx < len(self.predefined_seq_list_order), "seq_idx %i, expected predef num seqs %i" % (
          seq_idx, len(self.predefined_seq_list_order))
        expected_seq_name = self.predefined_seq_list_order[seq_idx]
        if expected_seq_name != segment_name:
          if segment_name in self.predefined_seq_list_order:
            raise Exception("seq_idx %i expected to be tag %r but got tag %r; tag %r is at idx %i" % (
              seq_idx, expected_seq_name, segment_name, segment_name,
              self.predefined_seq_list_order.index(segment_name)))
          raise Exception("seq_idx %i expected to be tag %r but got tag %r; tag %r not found" % (
            seq_idx, expected_seq_name, segment_name, segment_name))

      self.next_seq_to_be_added += 1
      self._num_timesteps += num_frames
      self.cond.notify_all()

      if seq_idx > self.requested_load_seq_end - 1 + self.SprintCachedSeqsMax:
        print("%s add_new_data: seq=%i, len=%i. Cache filled, waiting to get loaded..." % (
          self, seq_idx, num_frames), file=log.v5)
        while seq_idx > self.requested_load_seq_end - 1 + self.SprintCachedSeqsMin:
          assert not self.reached_final_seq
          assert seq_idx + 1 == self.next_seq_to_be_added
          self.cond.wait()

      self.added_data += [DatasetSeq(seq_idx, features, targets, seq_tag=segment_name)]
      self.cond.notify_all()
      return seq_idx

  def finish_sprint_epoch(self, seen_all=True):
    """
    Called by SprintInterface.getSegmentList().
    This is in a state where Sprint asks for the next segment after we just finished an epoch.
    Thus, any upcoming self.add_new_data() call will contain data from a segment in the new epoch.
    Thus, we finish the current epoch in Sprint.
    """
    with self.lock:
      self.reached_final_seq = True
      self.reached_final_seq_seen_all = seen_all
      self.ready_for_data = False
      self.cond.notify_all()

  def _shuffle_frames_in_seqs(self, start, end):
    assert False, "Shuffling in SprintDataset only via Sprint at the moment."

  def get_num_timesteps(self):
    """
    :rtype: int
    """
    with self.lock:
      assert self.reached_final_seq
      return self._num_timesteps

  @property
  def num_seqs(self):
    """
    :rtype: int
    """
    with self.lock:
      if self.predefined_seq_list_order:
        return len(self.predefined_seq_list_order)
      if not self.reached_final_seq:
        raise NotImplementedError
      return self.next_seq_to_be_added

  def have_seqs(self):
    """
    :rtype: bool
    """
    with self.lock:
      if self.next_seq_to_be_added > 0:
        return True
      self._wait_for_seq(0)
      return self.next_seq_to_be_added > 0

  def is_less_than_num_seqs(self, n):
    """
    :param int n:
    :rtype: bool
    """
    with self.lock:
      self._wait_for_seq(n)
      return n < self.next_seq_to_be_added

  def get_data_keys(self):
    """
    :rtype: list[str]
    """
    with self.lock:
      if not self.added_data:
        self._wait_for_seq(0)
      assert self.added_data
      return sorted(self.added_data[0].features.keys())

  def get_target_list(self):
    """
    :rtype: list[str]
    """
    keys = list(self.get_data_keys())
    if "data" in keys:
      keys.remove("data")
    return keys

  def set_complete_frac(self, frac):
    """
    :param float frac:
    """
    self._complete_frac = frac

  def get_complete_frac(self, seq_idx):
    """
    :param int seq_idx:
    :rtype: float
    """
    with self.lock:
      if self.predefined_seq_list_order:
        return float(seq_idx + 1) / len(self.predefined_seq_list_order)
      elif self._complete_frac is not None:
        if not self.next_seq_to_be_added:
          return self._complete_frac
        else:
          # We can do somewhat better. self._complete_frac is for self.next_seq_to_be_added.
          return self._complete_frac * float(seq_idx + 1) / self.next_seq_to_be_added
      else:
        return super(SprintDatasetBase, self).get_complete_frac(seq_idx)

  def get_seq_length(self, sorted_seq_idx):
    """
    :param int sorted_seq_idx:
    :rtype: Util.NumbersDict
    """
    with self.lock:
      self._wait_for_seq(sorted_seq_idx)
      return self._get_seq(sorted_seq_idx).num_frames

  def get_data(self, seq_idx, key):
    """
    :param int seq_idx:
    :param str key:
    :rtype: numpy.ndarray
    """
    with self.lock:
      self._wait_for_seq(seq_idx)
      return self._get_seq(seq_idx).features[key]

  def get_input_data(self, sorted_seq_idx):
    """
    :param int sorted_seq_idx:
    :rtype: numpy.ndarray
    """
    with self.lock:
      self._wait_for_seq(sorted_seq_idx)
      return self._get_seq(sorted_seq_idx).features["data"]

  def get_targets(self, target, sorted_seq_idx):
    """
    :param str target:
    :param int sorted_seq_idx:
    :rtype: numpy.ndarray
    """
    with self.lock:
      self._wait_for_seq(sorted_seq_idx)
      return self._get_seq(sorted_seq_idx).features.get(target, None)

  def get_ctc_targets(self, sorted_seq_idx):
    """
    :param int sorted_seq_idx:
    """
    assert False, "No CTC targets."

  def get_tag(self, sorted_seq_idx):
    """
    :param int sorted_seq_idx:
    :rtype: str
    """
    with self.lock:
      self._wait_for_seq(sorted_seq_idx)
      return self._get_seq(sorted_seq_idx).seq_tag


class ExternSprintDataset(SprintDatasetBase):
  """
  This is a Dataset which you can use directly in RETURNN.
  You can use it to get any type of data from Sprint (RWTH ASR toolkit),
  e.g. you can use Sprint to do feature extraction and preprocessing.

  This class is like SprintDatasetBase, except that we will start an external Sprint instance ourselves
  which will forward the data to us over a pipe.
  The Sprint subprocess will use SprintExternInterface to communicate with us.
  """

  # Do not change the argument names here, to not break existing configs.
  # noinspection PyPep8Naming
  def __init__(self, sprintTrainerExecPath, sprintConfigStr, partitionEpoch=None, **kwargs):
    """
    :param str|list[str] sprintTrainerExecPath:
    :param str | list[str] | ()->str | list[()->str] | ()->list[str] | ()->list[()->str] sprintConfigStr:
      via eval_shell_str
    :param int|None partitionEpoch: deprecated. use partition_epoch instead
    """
    super(ExternSprintDataset, self).__init__(**kwargs)
    self.add_data_thread_id = None
    self.sprint_trainer_exec_path = sprintTrainerExecPath
    self.sprint_config = sprintConfigStr
    if partitionEpoch:
      assert self.partition_epoch == 1, "don't provide partitionEpoch and partition_epoch"
      self.partition_epoch = partitionEpoch
    self._num_seqs = None
    self.child_pid = None  # type: typing.Optional[int]
    self.parent_pid = os.getpid()
    self.reader_thread = None  # type: typing.Optional[Thread]
    self.seq_list_file = None
    self.use_multiple_epochs()
    # There is no generic way to see whether Python is exiting.
    # This is our workaround. We check for it in self.run_inner().
    self.python_exit = False
    atexit.register(self._exit_handler)
    # We don't know about num_outputs yet, but we should.
    # Thus we call Sprint and immediately exit it.
    self._start_child(epoch=None, get_dim_only=True)

  def finish_epoch(self):
    """
    Called at the end of the epoch.
    """
    with self.lock:
      # Reset epoch such that exiting the child will go smoothly.
      super(ExternSprintDataset, self).init_seq_order(epoch=None, seq_list=None)
    # Exit child, before we overwrite anything, such as new epoch or seq_list.
    self._exit_child(wait_thread=True)
    super(ExternSprintDataset, self).finish_epoch()

  def _exit_handler(self):
    """
    Called at exit.
    """
    assert os.getpid() == self.parent_pid
    self.python_exit = True
    self._exit_child(wait_thread=False)

  def _exit_child(self, wait_thread=True):
    """
    :param bool wait_thread:
    """
    if self.child_pid:
      expected_exit_status = 0 if wait_thread and not self.python_exit else None
      if self._join_child(wait=False, expected_exit_status=expected_exit_status) is False:  # Not yet terminated.
        interrupt = not self.reached_final_seq_seen_all or not wait_thread
        if interrupt:
          print("%s: interrupt child proc %s" % (self, self.child_pid), file=log.v5)
          os.kill(self.child_pid, signal.SIGKILL)
          # Also join such that the process is cleaned up, and pipes get closed.
          self._join_child(wait=True, expected_exit_status=None)
          self.child_pid = None
      else:  # child process terminated
        self.child_pid = None
      if wait_thread and self.reader_thread:
        # Load all remaining data so that the reader thread is not waiting in self.add_new_data().
        while self.is_less_than_num_seqs(self.expected_load_seq_start + 1):
          if self.reached_final_seq:  # this is set by the reader thread
            break
          self.load_seqs(self.expected_load_seq_start + 1, self.expected_load_seq_start + 2)
        self.reader_thread.join()
        self.reader_thread = None
      try:
        self.pipe_p2c[1].close()
      except IOError:
        pass
      try:
        self.pipe_c2p[0].close()
      except IOError:
        pass
      if self.child_pid:
        self._join_child(wait=True, expected_exit_status=expected_exit_status)
        self.child_pid = None

  def _start_child(self, epoch, get_dim_only=False):
    """
    :param int|None epoch:
    :param bool get_dim_only:
    """
    assert self.child_pid is None
    assert self.reader_thread is None
    self.pipe_c2p = self._pipe_open()
    self.pipe_p2c = self._pipe_open()
    args = self._build_sprint_args()
    print("%s: epoch" % self, epoch, "exec", args, file=log.v5)

    pid = os.fork()
    if pid == 0:  # child
      # In case we are in some test environment or so, recover the original stdout/stderr.
      sys.stdin = sys.__stdin__
      sys.stdout = sys.__stdout__
      sys.stderr = sys.__stderr__
      from returnn.util import better_exchook
      better_exchook.install()
      # noinspection PyBroadException
      try:
        sys.stdin.close()  # Force no tty stdin.
        self.pipe_c2p[0].close()
        self.pipe_p2c[1].close()
        close_all_fds_except([0, 1, 2, self.pipe_c2p[1].fileno(), self.pipe_p2c[0].fileno()])
        os.execv(args[0], args)  # Does not return if successful.
        print("%s child exec failed." % self)
      except BaseException:
        print("%s child: Error when starting Sprint %r." % (self, args))
        sys.excepthook(*sys.exc_info())
      finally:
        print("%s child: exit" % self)
        # noinspection PyProtectedMember,PyUnresolvedReferences
        os._exit(1)
        return  # Not reached.

    # parent
    self.pipe_c2p[1].close()
    self.pipe_p2c[0].close()
    self.child_pid = pid

    try:
      init_signal, (input_dim, output_dim, num_segments) = self._read_next_raw()
      assert init_signal == b"init"
      assert isinstance(input_dim, int) and isinstance(output_dim, int)
      # Ignore num_segments. It can be totally different than the real number of sequences.
      self.set_dimensions(input_dim, output_dim)
    except Exception:
      print("%s: Sprint child process (%r) caused an exception." % (self, args), file=log.v1)
      sys.excepthook(*sys.exc_info())
      self._exit_child(wait_thread=False)
      raise Exception("%s Sprint init failed" % self)

    if get_dim_only:
      self._exit_child(wait_thread=False)

    else:
      self.reader_thread = Thread(target=self._reader_thread_proc, args=(pid, epoch),
                                  name="%s reader thread" % self)
      self.reader_thread.daemon = True
      self.reader_thread.start()

  # noinspection PyMethodMayBeStatic
  def _pipe_open(self):
    readend, writeend = os.pipe()
    if hasattr(os, "set_inheritable"):
      # Python 3 by default will close all fds in subprocesses. This will avoid that.
      os.set_inheritable(readend, True)
      os.set_inheritable(writeend, True)
    readend = os.fdopen(readend, "rb", 0)
    writeend = os.fdopen(writeend, "wb", 0)
    return readend, writeend

  @property
  def _my_python_mod_path(self):
    """
    :rtype: str
    """
    from returnn import __root_dir__
    return __root_dir__

  def _build_sprint_args(self):
    """
    :rtype: list[str]
    """
    config_str = "action:ExternSprintDataset,c2p_fd:%i,p2c_fd:%i" % (
      self.pipe_c2p[1].fileno(), self.pipe_p2c[0].fileno())
    if task_system.SharedMemNumpyConfig["enabled"]:
      config_str += ",EnableAutoNumpySharedMemPickling:True"
    epoch = self.returnn_epoch or 1
    assert epoch >= 1
    if isinstance(self.sprint_trainer_exec_path, (list, tuple)):
      args = list(self.sprint_trainer_exec_path)
    else:
      args = [self.sprint_trainer_exec_path]
    # First the user options. Usually also involves loading some config.
    args += eval_shell_str(self.sprint_config)
    # Now our options. They might overwrite some of the config settings. (That is why we do it after the user opts.)
    args += [
      "--*.seed=%i" % (self._get_random_seed_for_epoch(epoch=epoch) - 1)]
    if self.partition_epoch > 1:
      args += [
        "--*.corpus.partition=%i" % self.partition_epoch,
        "--*.corpus.select-partition=%i" % ((epoch - 1) % self.partition_epoch)]
    args += [
      "--*.python-segment-order=true",
      "--*.python-segment-order-pymod-path=%s" % self._my_python_mod_path,
      "--*.python-segment-order-pymod-name=returnn.sprint.extern_interface",
      "--*.use-data-source=false",
      "--*.trainer=python-trainer",
      "--*.pymod-path=%s" % self._my_python_mod_path,
      "--*.pymod-name=returnn.sprint.extern_interface",
      "--*.pymod-config=%s" % config_str]
    if self.predefined_seq_list_order:
      import tempfile
      self.seq_list_file = tempfile.mktemp(prefix="returnn-sprint-predefined-seq-list")
      with open(self.seq_list_file, "w") as f:
        for tag in self.predefined_seq_list_order:
          f.write(tag)
          f.write("\n")
        f.close()
      args += [
        "--*.corpus.segment-order-shuffle=false",
        "--*.corpus.segments.file=%s" % self.seq_list_file,
        "--*.corpus.segment-order=%s" % self.seq_list_file]
    if self.seq_tags_filter is not None:
      assert not self.predefined_seq_list_order
      import tempfile
      self.seq_list_file = tempfile.mktemp(prefix="returnn-sprint-predefined-seq-filter")
      with open(self.seq_list_file, "w") as f:
        for tag in self.seq_tags_filter:
          f.write(tag)
          f.write("\n")
        f.close()
      args += ["--*.corpus.segments.file=%s" % self.seq_list_file]
    return args

  def _read_next_raw(self):
    """
    :return: (data_type, args)
    :rtype: (str, object)
    """
    import struct
    size_raw = self.pipe_c2p[0].read(4)
    if len(size_raw) < 4:
      raise EOFError
    size, = struct.unpack("<i", size_raw)
    assert size > 0, "%s: We expect to get some non-empty package. Invalid Python mod in Sprint?" % (self,)
    stream = BytesIO()
    read_size = 0
    while read_size < size:
      data_raw = self.pipe_c2p[0].read(size - read_size)
      if len(data_raw) == 0:
        raise EOFError("%s: expected to read %i bytes but got EOF after %i bytes" % (self, size, read_size))
      read_size += len(data_raw)
      stream.write(data_raw)
    stream.seek(0)
    try:
      if PY3:
        # encoding is for converting Python2 strings to Python3.
        # Cannot use utf8 because Numpy will also encode the data as strings and there we need it as bytes.
        data_type, args = Unpickler(stream, encoding="bytes").load()
      else:
        data_type, args = Unpickler(stream).load()
    except EOFError:
      raise Exception("%s: parse error of %i bytes (%r)" % (self, size, stream.getvalue()))
    return data_type, args

  def _join_child(self, wait=True, expected_exit_status=None):
    """
    :param bool wait:
    :param int|None expected_exit_status:
    :return: whether the child has exited now
    :rtype: bool
    """
    assert self.child_pid
    options = 0 if wait else os.WNOHANG
    pid, exit_status = os.waitpid(self.child_pid, options)
    if not wait and pid == 0:
      return False
    assert pid == self.child_pid
    if expected_exit_status is not None:
      assert exit_status == expected_exit_status, "%s: Sprint exit code is %i" % (self, exit_status)
    return True

  def _reader_thread_proc(self, child_pid, epoch):
    """
    :param int child_pid:
    :param int epoch:
    """
    try:
      self.add_data_thread_id = thread.get_ident()

      self.init_sprint_epoch(epoch)
      have_seen_the_whole = False

      seq_count = 0
      while not self.python_exit and self.child_pid:
        try:
          data_type, args = self._read_next_raw()
        except (IOError, EOFError):
          with self.lock:
            if epoch != self.returnn_epoch:
              # We have passed on to a new epoch. This is a valid reason that the child has been killed.
              break
            if self.python_exit or not self.child_pid:
              break
          raise

        with self.lock:
          if epoch != self.returnn_epoch:
            break
          if self.python_exit or not self.child_pid:
            break

          if data_type == b"data":
            seq_count += 1
            segment_name, features, targets = args
            if segment_name is not None:
              segment_name = segment_name.decode("utf8")
            assert isinstance(features, numpy.ndarray)
            if isinstance(targets, dict):
              targets = {key.decode("utf8"): value for (key, value) in targets.items()}
            self.add_new_data(
              numpy_copy_and_set_unused(features),
              numpy_copy_and_set_unused(targets),
              segment_name=segment_name)
          elif data_type == b"exit":
            have_seen_the_whole = True
            break
          else:
            assert False, "not handled: (%r, %r)" % (data_type, args)

      if self.seq_list_file:
        try:
          os.remove(self.seq_list_file)
        except Exception as e:
          print("%s: error when removing %r: %r" % (self, self.seq_list_file, e), file=log.v5)
        finally:
          self.seq_list_file = None

      if not self.python_exit:
        with self.lock:
          self.finish_sprint_epoch(seen_all=have_seen_the_whole)
          if have_seen_the_whole:
            self._num_seqs = self.next_seq_to_be_added
      print("%s (proc %i) finished reading epoch %i, seen all %r (finished), num seqs %i" % (
        self, child_pid, epoch, have_seen_the_whole, seq_count), file=log.v5)

    except Exception as exc:
      if not self.python_exit:
        # Catch all standard exceptions.
        # Don't catch KeyboardInterrupt here because that will get send by the main thread
        # when it is exiting. It's never by the user because SIGINT will always
        # trigger KeyboardInterrupt in the main thread only.
        if epoch == self.returnn_epoch:
          with self.lock:
            self.finish_sprint_epoch(seen_all=False)
        try:
          print("%s reader failed (%s)" % (self, exc), file=log.v1)
          sys.excepthook(*sys.exc_info())
          print("")
        finally:
          # Exceptions are fatal. If we can recover, we should handle it in run_inner().
          interrupt_main()

  def init_seq_order(self, epoch=None, seq_list=None, seq_order=None):
    """
    :param int epoch:
    :param list[str]|None seq_list:
    :param list[int]|None seq_order:
    :rtype: bool
    """
    if seq_order is not None:
      raise NotImplementedError("Predefined sequence order via indices in ExternSprintDataset.")
    if seq_list is not None:
      assert self.partition_epoch == 1, "specifying partition_epoch and using seq_list not supported"
    if epoch is None:
      epoch = 1
    with self.lock:
      if (
            epoch == self.returnn_epoch and
            self.expected_load_seq_start == 0 and
            seq_list == self.predefined_seq_list_order):
        return
      # Reset epoch such that exiting the child will go smoothly.
      super(ExternSprintDataset, self).init_seq_order(epoch=None, seq_list=None, seq_order=None)
    # Exit child, before we overwrite anything, such as new epoch or seq_list.
    self._exit_child(wait_thread=True)
    with self.lock:  # Lock should not be needed now, but just to make it clean.
      if self._num_seqs:
        self._estimated_num_seqs = self._num_seqs  # last epoch num_seqs is a good estimate
      self._num_seqs = None  # we are not certain whether we have the same num_seqs for this epoch
      super(ExternSprintDataset, self).init_seq_order(epoch=epoch, seq_list=seq_list, seq_order=seq_order)
    self._start_child(epoch)
    return True


class SprintCacheDataset(CachedDataset2):
  """
  Can directly read Sprint cache files (and bundle files).
  Supports both cached features and cached alignments.
  For alignments, you need to provide all options for the AllophoneLabeling class, such as allophone file, etc.
  """

  class SprintCacheReader(object):
    """
    Helper class to read a Sprint cache directly.
    """
    def __init__(self, data_key, filename, data_type=None, allophone_labeling=None):
      """
      :param str data_key: e.g. "data" or "classes"
      :param str filename: to Sprint cache archive
      :param str|None data_type: "feat" or "align"
      :param dict[str] allophone_labeling: kwargs for :class:`AllophoneLabeling`
      """
      self.data_key = data_key
      from returnn.sprint.cache import open_file_archive
      self.sprint_cache = open_file_archive(filename)
      if not data_type:
        if data_key == "data":
          data_type = "feat"
        elif data_key == "classes":
          data_type = "align"
        else:
          # Some sensible defaults.
          if allophone_labeling:
            data_type = "align"
          else:
            data_type = "feat"
      assert data_type in ["feat", "align"]
      self.type = data_type
      self.allophone_labeling = None
      if allophone_labeling:
        from returnn.sprint.cache import AllophoneLabeling
        self.allophone_labeling = AllophoneLabeling(**allophone_labeling)
        self.sprint_cache.set_allophones(self.allophone_labeling.allophone_file)
      else:
        assert data_type != "align", "need allophone_labeling for 'align' type"
      self.content_keys = [fn for fn in self.sprint_cache.file_list() if not fn.endswith(".attribs")]
      if data_type == "align":
        self.num_labels = self.allophone_labeling.num_labels
        if self.num_labels < 2 ** 7:
          self.dtype = "int8"
        elif self.num_labels < 2 ** 15:
          self.dtype = "int16"
        else:
          assert self.num_labels < 2 ** 31
          self.dtype = "int32"
        self.num_dims = 1
        if self.allophone_labeling.state_tying_by_allo_state_idx:
          self.type = "align_raw"
      elif data_type == "feat":
        self.num_labels = self._get_feature_dim()
        self.num_dims = 2
        self.dtype = "float32"
      else:
        assert False

    def _get_feature_dim(self):
      """
      :rtype: int
      """
      assert self.type == "feat"
      assert self.content_keys
      times, feats = self.sprint_cache.read(self.content_keys[0], "feat")
      assert len(times) == len(feats) > 0
      feat = feats[0]
      assert isinstance(feat, numpy.ndarray)
      assert feat.ndim == 1
      return feat.shape[0]

    def read(self, name):
      """
      :param str name: content-filename for sprint cache
      :return: numpy array of shape (time, [num_labels])
      :rtype: numpy.ndarray
      """
      res = self.sprint_cache.read(name, typ=self.type)
      if self.type == "align":
        label_seq = numpy.array([self.allophone_labeling.get_label_idx(a, s) for (t, a, s) in res], dtype=self.dtype)
        assert label_seq.shape == (len(res),)
        return label_seq
      elif self.type == "align_raw":
        label_seq = numpy.array(
          [self.allophone_labeling.state_tying_by_allo_state_idx[a] for (t, a, s) in res], dtype=self.dtype)
        assert label_seq.shape == (len(res),)
        return label_seq
      elif self.type == "feat":
        times, feats = res
        assert len(times) == len(feats) > 0
        feat_mat = numpy.array(feats, dtype=self.dtype)
        assert feat_mat.shape == (len(times), self.num_labels)
        return feat_mat
      else:
        assert False

  def __init__(self, data, **kwargs):
    """
    :param dict[str,dict[str]] data: data-key -> dict which keys such as filename, see SprintCacheReader constructor
    """
    super(SprintCacheDataset, self).__init__(**kwargs)
    self.data = {key: self.SprintCacheReader(data_key=key, **opts) for (key, opts) in data.items()}
    self.seq_list_original = self.data["data"].content_keys
    self.seq_list_ordered = self.seq_list_original
    self._num_seqs = len(self.seq_list_original)
    self._check_matching_content_list()
    self.num_outputs = {key: (d.num_labels, d.num_dims) for (key, d) in self.data.items()}
    self.num_inputs = self.num_outputs["data"][0]
    self._seq_lens = None

  def _check_matching_content_list(self):
    data0 = self.data["data"]
    assert isinstance(data0, self.SprintCacheReader)
    sorted_list0 = sorted(data0.content_keys)
    for key, data in self.data.items():
      if key == "data":
        continue
      assert isinstance(data, self.SprintCacheReader)
      assert len(data.content_keys) == len(data0.content_keys)
      sorted_list1 = sorted(data.content_keys)
      for i in range(len(data.content_keys)):
        k0 = sorted_list0[i]
        k1 = sorted_list1[i]
        assert k0 == k1

  def init_seq_order(self, epoch=None, seq_list=None, seq_order=None):
    """
    :param int epoch:
    :param list[str]|None seq_list:
    :param list[int]|None seq_order:
    :rtype: bool
    """
    assert seq_list is None and seq_order is None
    need_reinit = self.epoch is None or self.epoch != epoch
    super(SprintCacheDataset, self).init_seq_order(epoch=epoch, seq_list=seq_list, seq_order=seq_order)
    self._num_seqs = len(self.seq_list_ordered)
    if not need_reinit:
      return False
    data0 = self.data["data"]
    assert isinstance(data0, self.SprintCacheReader)

    def get_seq_size(s):
      """
      :param int s:
      :rtype: int
      """
      return data0.sprint_cache.ft[self.seq_list_original[s]].size
    seq_index = self.get_seq_order_for_epoch(epoch, self.num_seqs, get_seq_len=get_seq_size)
    self.seq_list_ordered = [self.seq_list_original[s] for s in seq_index]
    self._num_seqs = len(self.seq_list_ordered)
    return True

  def get_dataset_seq_for_name(self, name, seq_idx=-1):
    """
    :param str name:
    :param int seq_idx:
    :rtype: DatasetSeq
    """
    data = {key: d.read(name) for (key, d) in self.data.items()}  # type: typing.Dict[str,numpy.ndarray]
    return DatasetSeq(seq_idx=seq_idx, seq_tag=name, features=data["data"], targets=data)

  def _collect_single_seq(self, seq_idx):
    """
    :type seq_idx: int
    :rtype: DatasetSeq | None
    :returns DatasetSeq or None if seq_idx >= num_seqs.
    """
    if seq_idx >= self.num_seqs:
      return None
    seq_tag = self.get_tag(seq_idx)  # type: str
    return self.get_dataset_seq_for_name(seq_idx=seq_idx, name=seq_tag)

  def get_data_keys(self):
    """
    :rtype: list[str]
    """
    return self.data.keys()

  def get_target_list(self):
    """
    :rtype: list[str]
    """
    return [key for (key, d) in self.data.items() if d.type == "align"]

  def get_tag(self, sorted_seq_idx):
    """
    :rtype: str
    """
    return self.seq_list_ordered[sorted_seq_idx]


def demo():
  """
  Demo.
  """
  print("SprintDataset demo.")
  from argparse import ArgumentParser
  from returnn.util.basic import progress_bar_with_time
  from returnn.log import log
  from returnn.config import Config
  from returnn.datasets.basic import init_dataset
  arg_parser = ArgumentParser()
  arg_parser.add_argument("--config", help="config with ExternSprintDataset", required=True)
  arg_parser.add_argument("--sprint_cache_dataset", help="kwargs dict for SprintCacheDataset", required=True)
  arg_parser.add_argument("--max_num_seqs", default=sys.maxsize, type=int)
  arg_parser.add_argument("--action", default="compare", help="compare or benchmark")
  args = arg_parser.parse_args()
  log.initialize(verbosity=[4])
  sprint_cache_dataset_kwargs = eval(args.sprint_cache_dataset)
  assert isinstance(sprint_cache_dataset_kwargs, dict)
  sprint_cache_dataset = SprintCacheDataset(**sprint_cache_dataset_kwargs)
  print("SprintCacheDataset: %r" % sprint_cache_dataset)
  config = Config()
  config.load_file(args.config)
  dataset = init_dataset(config.typed_value("train"))
  print("Dataset via config: %r" % dataset)
  assert sprint_cache_dataset.num_inputs == dataset.num_inputs
  assert tuple(sprint_cache_dataset.num_outputs["classes"]) == tuple(dataset.num_outputs["classes"])
  sprint_cache_dataset.init_seq_order(epoch=1)

  if args.action == "compare":
    print("Iterating through dataset...")
    seq_idx = 0
    dataset.init_seq_order(epoch=1)
    while seq_idx < args.max_num_seqs:
      if not dataset.is_less_than_num_seqs(seq_idx):
        break
      dataset.load_seqs(seq_idx, seq_idx + 1)
      tag = dataset.get_tag(seq_idx)
      assert not tag.startswith("seq-"), "dataset does not provide tag-names for seqs"
      dataset_seq = sprint_cache_dataset.get_dataset_seq_for_name(tag)
      data = dataset.get_data(seq_idx, "data")
      targets = dataset.get_data(seq_idx, "classes")
      assert data.shape == dataset_seq.features["data"].shape
      assert targets.shape == dataset_seq.features["classes"].shape
      assert numpy.allclose(data, dataset_seq.features["data"])
      assert numpy.allclose(targets, dataset_seq.features["classes"])
      seq_idx += 1
      progress_bar_with_time(dataset.get_complete_frac(seq_idx))

    print("Finished through dataset. Num seqs: %i" % seq_idx)
    print("SprintCacheDataset has num seqs: %i." % sprint_cache_dataset.num_seqs)

  elif args.action == "benchmark":
    print("Iterating through dataset...")
    start_time = time.time()
    seq_tags = []
    seq_idx = 0
    dataset.init_seq_order(epoch=1)
    while seq_idx < args.max_num_seqs:
      if not dataset.is_less_than_num_seqs(seq_idx):
        break
      dataset.load_seqs(seq_idx, seq_idx + 1)
      tag = dataset.get_tag(seq_idx)
      assert not tag.startswith("seq-"), "dataset does not provide tag-names for seqs"
      seq_tags.append(tag)
      dataset.get_data(seq_idx, "data")
      dataset.get_data(seq_idx, "classes")
      seq_idx += 1
      progress_bar_with_time(dataset.get_complete_frac(seq_idx))
    print("Finished through dataset. Num seqs: %i, time: %f" % (seq_idx, time.time() - start_time))
    print("SprintCacheDataset has num seqs: %i." % sprint_cache_dataset.num_seqs)
    if hasattr(dataset, "exit_handler"):
      dataset.exit_handler()
    else:
      print("No way to stop any background tasks.")
    del dataset

    start_time = time.time()
    print("Iterating through SprintCacheDataset...")
    for i, tag in enumerate(seq_tags):
      sprint_cache_dataset.get_dataset_seq_for_name(tag)
      progress_bar_with_time(float(i) / len(seq_tags))
    print("Finished through SprintCacheDataset. time: %f" % (time.time() - start_time,))

  else:
    raise Exception("invalid action: %r" % args.action)


if __name__ == "__main__":
  from returnn.util import better_exchook
  better_exchook.install()
  try:
    demo()
  except KeyboardInterrupt:
    print("KeyboardInterrupt.")
    sys.exit(1)




import sys

import _setup_test_env  # noqa
import unittest
from nose.tools import assert_equal, assert_is_instance, assert_in, assert_true, assert_false
from returnn.network_description import LayerNetworkDescription
from returnn.config import Config
from returnn.util.basic import obj_diff_str
from pprint import pprint
from returnn.util import better_exchook
from returnn.util.basic import BackendEngine

try:
  # noinspection PyPackageRequirements
  import theano
  BackendEngine.select_engine(engine=BackendEngine.Theano)
except ImportError:
  theano = None


if theano:
  import returnn.theano.util

  returnn.theano.util.monkey_patches()

  from returnn.theano.network import LayerNetwork
  from returnn.theano.layers.hidden import ForwardLayer

else:
  LayerNetwork = None
  ForwardLayer = None


def test_init():
  n_in = 5
  n_out = {"classes": (10, 1)}
  desc = LayerNetworkDescription(
    num_inputs=n_in, num_outputs=n_out,
    hidden_info=[],
    output_info={},
    default_layer_info={})

  assert_equal(desc.num_inputs, n_in)
  assert_equal(desc.num_outputs, n_out)


def test_num_inputs_outputs_old():
  n_in = 5
  n_out = 10
  config = Config()
  config.update({"num_inputs": n_in, "num_outputs": n_out})
  num_inputs, num_outputs = LayerNetworkDescription.num_inputs_outputs_from_config(config)
  assert_equal(num_inputs, n_in)
  assert_is_instance(num_outputs, dict)
  assert_equal(len(num_outputs), 1)
  assert_in("classes", num_outputs)
  assert_equal(num_outputs["classes"], (n_out, 1))


def test_num_inputs_outputs_special_dataset():
  config = Config()
  config.update({
    "train": {"class": "CopyTaskDataset", "num_seqs": 1000, "nsymbols": 80, "minlen": 100, "maxlen": 100},
    "num_outputs": {"data": [80, 1], "classes": [80, 1]}})
  num_inputs, num_outputs = LayerNetworkDescription.num_inputs_outputs_from_config(config)
  assert_equal(num_inputs, 80)
  assert_in("data", num_outputs)
  assert_in("classes", num_outputs)
  assert_equal(num_outputs["classes"], (80, 1))
  assert_equal(num_outputs["data"], (80, 1))


config1_dict = {
  "num_inputs": 5,
  "num_outputs": 10,
  "hidden_size": (7, 8,),
  "hidden_type": "forward",
  "activation": "relu",
  "bidirectional": False,
}

config2_dict = {
  "use_theano": True,
  "pretrain": "default",
  "num_inputs": 40,
  "num_outputs": 4498,
  "bidirectional": True,
  "hidden_size": (500, 500, 500),
  "hidden_type": "lstm_opt",
  "activation": "sigmoid",
  "dropout": 0.1,
}



def test_config1_basic():
  config = Config()
  config.update(config1_dict)
  desc = LayerNetworkDescription.from_config(config)
  assert_is_instance(desc.hidden_info, list)
  assert_equal(len(desc.hidden_info), len(config1_dict["hidden_size"]))
  assert_equal(desc.num_inputs, config1_dict["num_inputs"])


@unittest.skipIf(not theano, "not theano")
def test_network_config1_init():
  config = Config()
  config.update(config1_dict)
  network = LayerNetwork.from_config_topology(config)
  assert_in("hidden_0", network.hidden)
  assert_in("hidden_1", network.hidden)
  assert_equal(len(network.hidden), 2)
  assert_is_instance(network.hidden["hidden_0"], ForwardLayer)
  assert_equal(network.hidden["hidden_0"].layer_class, "hidden")
  assert_false(network.recurrent)

  json_content = network.to_json_content()
  pprint(json_content)
  assert_in("hidden_0", json_content)
  assert_equal(json_content["hidden_0"]["class"], "hidden")
  assert_in("hidden_1", json_content)
  assert_in("output", json_content)


@unittest.skipIf(not theano, "not theano")
def test_NetworkDescription_to_json_config1():
  config = Config()
  config.update(config1_dict)
  desc = LayerNetworkDescription.from_config(config)
  desc_json_content = desc.to_json_content()
  pprint(desc_json_content)
  assert_in("hidden_0", desc_json_content)
  assert_equal(desc_json_content["hidden_0"]["class"], "forward")
  assert_in("hidden_1", desc_json_content)
  assert_in("output", desc_json_content)
  orig_network = LayerNetwork.from_description(desc)
  assert_in("hidden_0", orig_network.hidden)
  assert_in("hidden_1", orig_network.hidden)
  assert_equal(len(orig_network.hidden), 2)
  assert_is_instance(orig_network.hidden["hidden_0"], ForwardLayer)
  assert_equal(orig_network.hidden["hidden_0"].layer_class, "hidden")
  orig_json_content = orig_network.to_json_content()
  pprint(orig_json_content)
  assert_in("hidden_0", orig_json_content)
  assert_equal(orig_json_content["hidden_0"]["class"], "hidden")
  assert_in("hidden_1", orig_json_content)
  assert_in("output", orig_json_content)
  new_network = LayerNetwork.from_json(
    desc_json_content,
    config1_dict["num_inputs"],
    {"classes": (config1_dict["num_outputs"], 1)})
  new_json_content = new_network.to_json_content()
  if orig_json_content != new_json_content:
    print(obj_diff_str(orig_json_content, new_json_content))
    assert_equal(orig_json_content, new_network.to_json_content())


@unittest.skipIf(not theano, "not theano")
def test_config1_to_json_network_copy():
  config = Config()
  config.update(config1_dict)
  orig_network = LayerNetwork.from_config_topology(config)
  orig_json_content = orig_network.to_json_content()
  pprint(orig_json_content)
  new_network = LayerNetwork.from_json(orig_json_content, orig_network.n_in, orig_network.n_out)
  assert_equal(orig_network.n_in, new_network.n_in)
  assert_equal(orig_network.n_out, new_network.n_out)
  new_json_content = new_network.to_json_content()
  if orig_json_content != new_json_content:
    print(obj_diff_str(orig_json_content, new_json_content))
    assert_equal(orig_json_content, new_network.to_json_content())


@unittest.skipIf(not theano, "not theano")
def test_config2_bidirect_lstm():
  config = Config()
  config.update(config2_dict)
  desc = LayerNetworkDescription.from_config(config)
  assert_true(desc.bidirectional)
  network = LayerNetwork.from_config_topology(config)
  net_json = network.to_json_content()
  pprint(net_json)
  assert_in("output", net_json)
  assert_in("hidden_0_fw", net_json)
  assert_in("hidden_0_bw", net_json)
  assert_in("hidden_1_fw", net_json)
  assert_in("hidden_1_bw", net_json)
  assert_in("hidden_2_fw", net_json)
  assert_in("hidden_2_bw", net_json)
  assert_equal(net_json["output"]["from"], ["hidden_2_fw", "hidden_2_bw"])
  assert_equal(len(net_json), 7)


if __name__ == "__main__":
  better_exchook.install()
  if len(sys.argv) <= 1:
    for k, v in sorted(globals().items()):
      if k.startswith("test_"):
        print("-" * 40)
        print("Executing: %s" % k)
        try:
          v()
        except unittest.SkipTest as exc:
          print("SkipTest:", exc)
        print("-" * 40)
    print("Finished all tests.")
  else:
    assert len(sys.argv) >= 2
    for arg in sys.argv[1:]:
      print("Executing: %s" % arg)
      if arg in globals():
        globals()[arg]()  # assume function and execute
      else:
        eval(arg)  # assume Python code and execute



"""
This provides the main :func:`import_` function.
"""

import types
import importlib
from .git import get_repo_path
from .common import module_name


def import_(repo, path, version=None):
  """
  :param str repo: e.g. "github.com/rwth-i6/returnn-experiments"
  :param str path: path inside the repo, without starting "/"
  :param str|None version: e.g. "20211231-0123abcd0123". None for development working copy
  :rtype: object|types.ModuleType
  """
  assert path and path[:1] != "/" and ".." not in path
  repo_path = get_repo_path(repo=repo, version=version)
  # `module_name` has the side effect that `import_module` below will just work.
  mod_name = module_name(repo=repo, repo_path=repo_path, path=path, version=version)
  return importlib.import_module(mod_name)




from __future__ import print_function

import _setup_test_env  # noqa
from subprocess import Popen, PIPE, STDOUT, CalledProcessError
import re
import os
import sys
from glob import glob
import unittest
from nose.tools import assert_less
from returnn.util import better_exchook


py = sys.executable
print("Python:", py)

os.chdir((os.path.dirname(__file__) or ".") + "/..")
assert os.path.exists("rnn.py")


def build_env(env_update=None):
  """
  :param dict[str,str]|None env_update:
  :return: env dict for Popen
  :rtype: dict[str,str]
  """
  theano_flags = {key: value for (key, value)
                  in [s.split("=", 1) for s in os.environ.get("THEANO_FLAGS", "").split(",") if s]}
  # First set some sane default for compile dir.
  theano_flags.setdefault("compiledir_format",
                          "compiledir_%(platform)s-%(processor)s-%(python_version)s-%(python_bitwidth)s")
  # Extend compile dir for our tests in a subprocess.
  # This is important because there might be other tests where we import Theano in this
  # process and any subprocess would block because the parent process has locked the compile-dir.
  theano_flags["compiledir_format"] += "--nosetests"
  # Nose-tests will set mode=FAST_COMPILE. We don't want this for our tests as it is way too slow.
  theano_flags["mode"] = "FAST_RUN"
  env_update_ = os.environ.copy()
  env_update_["THEANO_FLAGS"] = ",".join(["%s=%s" % (key, value) for (key, value) in theano_flags.items()])
  if env_update:
    env_update_.update(env_update)
  return env_update_


def run(*args, env_update=None):
  args = list(args)
  print("run:", args)
  # RETURNN by default outputs on stderr, so just merge both together
  p = Popen(args, stdout=PIPE, stderr=STDOUT, env=build_env(env_update=env_update))
  out, _ = p.communicate()
  if p.returncode != 0:
    print("Return code is %i" % p.returncode)
    print("std out/err:\n---\n%s\n---\n" % out.decode("utf8"))
    raise CalledProcessError(cmd=args, returncode=p.returncode, output=out)
  return out.decode("utf8")


def run_and_parse_last_fer(*args, **kwargs):
  out = run(*args, **kwargs)
  parsed_fer = None
  for line in out.splitlines():
    # example: epoch 5 score: 0.0231807245472 elapsed: 0:00:04 dev: score 0.0137521058997 error 0.00268961807423
    m = re.match("epoch [0-9]+ score: .* dev: .* error ([0-9.]+)\\s?", line)
    if not m:
      # example: dev: score 0.03350000149202181 error 0.009919877954075871
      m = re.match("dev: score .* error ([0-9.]+)\\s?", line)
    if not m:
      continue
    parsed_fer = float(m.group(1))
  err_msg = "ERROR: No epoch dev errors found in output"
  assert parsed_fer is not None, "%s.\nOutput:\n\n%s\n\n%s." % (err_msg, out, err_msg)
  return parsed_fer


def run_config_get_fer(config_filename, env_update=None):
  cleanup_tmp_models(config_filename)
  fer = run_and_parse_last_fer(
    py, "rnn.py", config_filename, "++log_verbosity", "5", env_update=env_update)
  print("FER:", fer)
  cleanup_tmp_models(config_filename)
  return fer


def cleanup_tmp_models(config_filename):
  assert os.path.exists(config_filename)
  from returnn.config import Config
  config = Config()
  config.load_file(config_filename)
  model_filename = config.value('model', '')
  assert model_filename
  # Remove existing models
  assert model_filename.startswith("/tmp/")
  for f in glob(model_filename + ".*"):
    os.remove(f)


def test_demo_theano_task12ax():
  fer = run_config_get_fer("demos/demo-theano-task12ax.config")
  assert_less(fer, 0.01)


if __name__ == '__main__':
  better_exchook.install()
  import sys
  fns = sys.argv[1:]
  if not fns:
    fns = [arg for arg in globals() if arg.startswith("test_")]
  for arg in fns:
    f = globals()[arg]
    print("-" * 20)
    print("Run:", f)
    print("-" * 20)
    try:
      f()
    except unittest.SkipTest as exc:
      print("(SkipTest: %s)" % exc)
    print("-" * 20)
    print("Ok.")
    print("-" * 20)
